/* Copyright 2003,2004 Geoff Archer
 *
 * This file is part of ForceRemap.
 *
 * ForceRemap is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ForceRemap is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ForceRemap; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/////////////////////////////////////////////////////////////////////////////////////////// Defines

// compiler options
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN

#define RGB_16_TO_24 255.0/248.0

#define MAXVAL    1000.0
#define MAXSCALE  20.0

#define WM_USERCLOSE WM_USER+1
#define WM_UPDATE WM_USER+2

// the universal bitmap ID
#define BITMAP_ID 0x4D42

#define PI 3.141592654

// main window dimensions
#define SEP          8
#define SB           10
#define BUTTON       (2*SB)+1

// ID's for user commands
#define ID_MENUOPEN           101
#define ID_MENUREVERT         102
#define ID_MENUCLOSE          103
#define ID_MENUSAVE           104
#define ID_MENUSAVEAS         105
#define ID_MENUOPTIONS        106
#define ID_MENUHIDECONTROLS   107
#define ID_MENUABOUT          108
#define ID_MENUOFFSET         109

#define ID_GLV                201
#define ID_MOUSELOOK          202
#define ID_RESETVIEW          203

#define ID_MESHLB             301
#define ID_MESHHIDE           302
#define ID_MESHWIRE           303
#define ID_MESHSHOW           304
#define ID_MESHHIDEALL        305
#define ID_MESHWIREALL        306
#define ID_MESHSHOWALL        307

#define ID_GROUPLB            401
#define ID_GROUPSELECTNONE    402
#define ID_GROUPCHANGEBMP     403

#define ID_UVORTHOFRONT       501
#define ID_UVORTHOBACK        502
#define ID_UVORTHOTOP         503
#define ID_UVORTHOBOTTOM      504
#define ID_UVORTHOLEFT        505
#define ID_UVORTHORIGHT       506
#define ID_UVEDIT             507
#define ID_UVHELP             508
#define ID_UVSELECTALL        509
#define ID_UVSELECTNONE       510
#define ID_UVSELECTINVERT     511
#define ID_UVSCALE            512
#define ID_UVAMOUNT           513
#define ID_UVXAXIS            514
#define ID_UVYAXIS            515
#define ID_UVBOTHAXES         516

#define ID_OPTIONSMATFIRST    600
#define ID_OPTIONSBMPFIRST    601
#define ID_OPTIONSMATSECOND   602
#define ID_OPTIONSBMPSECOND   603
#define ID_OPTIONSCMPBROWSE   604
#define ID_OPTIONSCMPDEFAULT  605
#define ID_OPTIONSOK          606
#define ID_OPTIONSCANCEL      607

////////////////////////////////////////////////////////////////////////////////////////// Includes

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <fstream.h>
#include <math.h>
#include <winerror.h>

#include <gl/gl.h>
#include <gl/glu.h> // OpenGL utilities

//////////////////////////////////////////////////////////////////////////////////////// Prototypes

class Camera;
class C3DOModel;
class CUVMap;
class CMaterial;
class CMatrix;
class CMatrixNode;
class CCmpPalette;

void FileDlg_Init();
BOOL FileDlg_GetOpenPathname(HWND, PSTR, PSTR, char *);
BOOL FileDlg_GetSavePathname(HWND, PSTR, PSTR, char *);

void GL_ClearGLV();
void GL_InitializeOrthoView();

bool AllWindows_Create();
void AllWindows_Close();

void Menu_Enable(bool);

void ProgramMode1(void);
void ProgramMode1to2(void);
void ProgramMode2to3(bool);
void ProgramMode3to4(bool);

void Command_Open(char *);
void Command_Close();
void Command_Save();
void Command_SaveAs(char *);

void Support_SetGlobalFilenameInfo(char *);
void Support_SetAppCaption();

void Options_LoadFromReg();
void Options_SaveToReg();

void Msg(HWND, char *, char *, ...);

/////////////////////////////////////////////////////////////////////////////////////////// Globals

HINSTANCE  ghinstApp;

const char gszAppName           [] = "ForceRemap 1.1";
const char gszMainClassName     [] = "MainWindowClass";
const char gszControlsClassName [] = "ControlsWindowClass";
const char gszGLVClassName      [] = "GLViewportClass";
const char gszOptionsClassName  [] = "OptionsWindowClass";

int  giProgramMode;
bool gbControlPanelShow;
bool gbOptionsShow;

bool gMatFirst;
bool gBmpFirst;
bool gMatSecond;
bool gBmpSecond;
const unsigned char gszDfltCmp[] = "Default (no file necessary)\0";
unsigned char gCmpPath[_MAX_PATH];
bool gCmpHasBeenLoaded;

Camera      *gCamera;
C3DOModel   *gmodel;          // current 3DO model
CUVMap      *guvmap;          // current UV map
CCmpPalette *gPalette;

bool  gbModelMode_Active;     // true if the user is viewing a 3DO model
bool  gbUVMapMode_Active;     // true if the user is editing a UV map

int   giMouseX, giMouseY;     // global mouse coordinates
bool  gbKeyPressed[256];      // holds true for keys that are pressed

char  gszEmptyString     [1];
char  gszProjectPathname [_MAX_PATH];
char  gszProjectFolder   [_MAX_DIR];
char  gszProjectTitlename[_MAX_FNAME];

//////////////

HWND ghwndOptions;
HWND ghwndOptionsBox1;
HWND ghwndOptionsMatFirst;
HWND ghwndOptionsBmpFirst;
HWND ghwndOptionsMatSecond;
HWND ghwndOptionsBmpSecond;
HWND ghwndOptionsBox3;
HWND ghwndOptionsCmpFile1;
HWND ghwndOptionsCmpFile2;
HWND ghwndOptionsCmpBrowse;
HWND ghwndOptionsDefault;
HWND ghwndOptionsOK;
HWND ghwndOptionsCancel;

//////////////

HWND  ghwndMain;
HMENU ghmenuMain;

HWND  ghwndScrollZoom;
HWND  ghwndScrollPitch;
HWND  ghwndScrollYaw;
HWND  ghwndScrollPanY;
HWND  ghwndScrollPanX;
HWND  ghwndMouseLook;
HWND  ghwndResetView;

HWND  ghwndGLV;
HDC   ghdcGLV;
HGLRC ghrcGLV;
RECT  grectGLV;

//////////////

HWND  ghwndControls;

//////////////

HWND  ghwndMeshBox;
HWND  ghwndMeshLB;
HWND  ghwndMeshHide;
HWND  ghwndMeshWire;
HWND  ghwndMeshShow;
HWND  ghwndMeshHideAll;
HWND  ghwndMeshWireAll;
HWND  ghwndMeshShowAll;

//////////////

HWND  ghwndGroupBox;
HWND  ghwndGroupLB;
HWND  ghwndGroupSelectNone;
HWND  ghwndGroupChangeBmp;

//////////////

HWND  ghwndUVBox;
HWND  ghwndUVOrthoTop;
HWND  ghwndUVOrthoBottom;
HWND  ghwndUVOrthoFront;
HWND  ghwndUVOrthoBack;
HWND  ghwndUVOrthoLeft;
HWND  ghwndUVOrthoRight;

HWND  ghwndUVBox2;
HWND  ghwndUVEdit;
HWND  ghwndUVHelp;
HWND  ghwndUVStaticSel;
HWND  ghwndUVSelectAll;
HWND  ghwndUVSelectNone;
HWND  ghwndUVSelectInvert;
HWND  ghwndUVScale;
HWND  ghwndUVScaleAmount;
HWND  ghwndUVStaticOn;
HWND  ghwndUVScaleX;
HWND  ghwndUVScaleY;
HWND  ghwndUVScaleBoth;

///////////////////////////////////////////////////////////////////////////////////////// Type defs

enum GL_rendermode
	{
	HIDE,
	WIRE,
	SHOW
	};

//////////////////////////////////////////////////////////////////////////////////////// Class defs

class Point3
	{
	public:

	double x, y, z;

	Point3()
		{x = y = z = 0.0;}

	Point3(double xx, double yy, double zz)
		{x = xx; y = yy; z = zz;}

	void set(double xx, double yy, double zz)
		{x = xx; y = yy; z = zz;}

	void set(Point3& p)
		{x = p.x; y = p.y; z = p.z;}

	void add(double xx, double yy, double zz)
		{ x += xx; y += yy; z += zz; }

	void mult(double m)
		{ x *= m; y *= m; z *= m; }
	};

class Vector3
	{
	public:

	double x, y, z;

	Vector3()
		{x = y = z = 0.0;}

	Vector3(double xx, double yy, double zz)
		{x = xx; y = yy; z = zz;}

	Vector3(Vector3& v)
		{x = v.x; y = v.y; z = v.z;}

	void set(double dx, double dy, double dz)
		{ x = dx; y = dy; z = dz;}

	void set(Vector3 v)
		{ x = v.x; y = v.y; z = v.z;}

	void scalarMult(double m)
		{ x *= m; y *= m; z *= m; }

	double dot(Vector3 b) // return this dotted with b
		{return x * b.x + y * b.y + z * b.z;}

	Vector3 cross(Vector3 b) //return this cross b
		{
		Vector3 c(y*b.z - z*b.y, z*b.x - x*b.z, x*b.y - y*b.x);
		return c;
		}

	void flip()
		{x = -x; y = -y; z = -z;} // reverse this vector

	void setDiff(Point3& a, Point3& b) // set to difference a - b
		{ x = a.x - b.x; y = a.y - b.y; z = a.z - b.z;}

	void normalize() // adjust this vector to unit length
		{
		double sizeSq = x * x + y * y + z * z;

		if(sizeSq < 0.0000001)
			return; // does nothing to zero vectors

		double scaleFactor = 1.0/sqrt(sizeSq);
		x *= scaleFactor; y *= scaleFactor; z *= scaleFactor;
		}
	};

class Camera
	{
 private:

	Point3 eye, center; Vector3 up;

	double hAngle, vAngle;
	int    Xscrollpos, Yscrollpos, Zoompos;

	double panXfactor, panYfactor;
	double zoomFactor, conversionFactor;

	bool   bMouseLook;

 public:

	double getVAngle()      const { return vAngle; }
	double getHAngle()      const { return hAngle; }
	int    getXscrollpos()  const { return Xscrollpos; }
	int    getYscrollpos()  const { return Yscrollpos; }
	int    getZoompos()     const { return Zoompos; }
	double getScaleFactor() const { return zoomFactor * conversionFactor; }
	bool   getMouseLook()   const { return bMouseLook; }

	void ResetView(double radius)
		{
		eye.set(0.0, 0.0, -MAXVAL * MAXSCALE);
		center.set(0.0, 0.0, 0.0);
		up.set(0.0, 1.0, 0.0);

		hAngle = vAngle = 0.0;

		panXfactor = 200.0 / (radius * 2.0);
		panYfactor = 200.0 / (radius * 2.0);

		Xscrollpos = Yscrollpos = Zoompos = 0;

		conversionFactor = MAXVAL / radius;
		zoomFactor = 1.0;

		bMouseLook = false;
		SetWindowText(ghwndMouseLook, "M");

		SetScrollPos(ghwndScrollPitch, SB_CTL, 0, true);
		SetScrollPos(ghwndScrollYaw,   SB_CTL, 0, true);
		SetScrollPos(ghwndScrollZoom,  SB_CTL, 0, true);
		SetScrollPos(ghwndScrollPanX,  SB_CTL, 0, true);
		SetScrollPos(ghwndScrollPanY,  SB_CTL, 0, true);
		}

	void doLookAt()
		{ gluLookAt(eye.x,eye.y,eye.z, center.x,center.y,center.z, up.x,up.y,up.z); }

	void AngleV(int amount, bool clamp)
		{
		vAngle += amount;
		if (vAngle > 180.0)
			if (clamp) vAngle = 180.0;
			else vAngle -= 360.0;
		else if (vAngle < -180.0)
			if (clamp) vAngle = -180.0;
			else vAngle += 360.0;
		SetScrollPos(ghwndScrollPitch, SB_CTL, -(int)vAngle, true);
		}

	void AngleH(int amount, bool clamp)
		{
		hAngle += amount;
		if (hAngle > 180.0)
			if (clamp) hAngle = 180.0;
			else hAngle -= 360.0;
		else if (hAngle < -180.0)
			if (clamp) hAngle = -180.0;
			else hAngle += 360.0;
		SetScrollPos(ghwndScrollYaw, SB_CTL, (int)hAngle, true);
		}

	void PanX(int amount)
		{
		Xscrollpos += amount;
		if (Xscrollpos > 100) Xscrollpos = 100;
		else if (Xscrollpos < -100) Xscrollpos = -100;
		SetScrollPos(ghwndScrollPanX, SB_CTL, Xscrollpos, true);

		center.x = eye.x = -(double)Xscrollpos / panXfactor * zoomFactor * conversionFactor;
		}

	void PanY(int amount)
		{
		Yscrollpos += amount;
		if (Yscrollpos > 100) Yscrollpos = 100;
		else if (Yscrollpos < -100) Yscrollpos = -100;
		SetScrollPos(ghwndScrollPanY, SB_CTL, Yscrollpos, true);

		center.y = eye.y = -(double)Yscrollpos / panYfactor * zoomFactor * conversionFactor;
		}

	void Zoom(int amount)
		{
		Zoompos += amount;
		if (Zoompos < -100) Zoompos = -100;
		else if (Zoompos > 100) Zoompos = 100;
		SetScrollPos(ghwndScrollZoom, SB_CTL, -Zoompos, true);

		zoomFactor = pow(1.03, Zoompos);

		center.x = eye.x = -(double)Xscrollpos / panXfactor * zoomFactor * conversionFactor;
		center.y = eye.y = -(double)Yscrollpos / panYfactor * zoomFactor * conversionFactor;
		}

	void ToggleMouseLook()
		{
		bMouseLook = !bMouseLook;

		if (bMouseLook)
			SetWindowText(ghwndMouseLook, "m");
		else
			SetWindowText(ghwndMouseLook, "M");
		}
	};

///////////////////////////////////////////////////////////////////////////////////////////////////

class JK_offset
	{
	public:
	double x, y, z;
	};

class GL_offset
	{
	public:
	double x, y, z;
	};

class JK_normal
	{
	public:
	double x, y, z;
	};

class JK_vertex
	{
	public:
	double x, y, z, i;
	int temp;   // misc variable used during algorithms
	};

class GL_vertex
	{
	public:
	double x, y, z;
	bool flag;  // misc variable used during algorithms
	};

class JK_txvert
	{
	public:
	double s, t;
	};

class GL_txvert
	{
	public:
	double s, t;
	bool is_selected;

	bool flag;  // misc variable used during algorithms
	int redirect;
	};

class JK_pair
	{
	public:
	int vertexID, txvertID;
	};

class GL_pair
	{
	public:
	int vertexID, txvertID;
	};

class JK_face
	{
	public:

	int matID;
	int groupID;

	char jkType[64];
	int jkGeo, jkLight, jkTex;
	double jkExtralight;

	int numjkPairs;
	JK_pair *jkPairs;

	int jkFaceID;
	};

class GL_face
	{
	public:

	int numglPairs;
	GL_pair *glPairs;

	int jkFaceID;
	};

class GL_group
	{
	public:

	char glGroupName[_MAX_FNAME];
	int matID;

	int numglTxverts;
	GL_txvert *glTxverts;

	int numglFaces;
	GL_face *glFaces;

	bool is_selected;
	bool is_original;
	};

///////////////////////////////////////////////////////////////////////////////////////////////////

class CMatrixNode
	{
	public:

	int value;
	CMatrixNode* right;
	CMatrixNode* lower;
	CMatrixNode* left;
	CMatrixNode* upper;

	CMatrixNode();
	};

CMatrixNode::CMatrixNode()
	{
	value = 0;
	right = NULL;
	lower = NULL;
	left  = NULL;
	upper = NULL;
	}

class CMatrix
	{
	public:

	int numrows, numcols;
	CMatrixNode *head;

	CMatrix(int, int);
	~CMatrix();

	Set(int, int, int);
	int Get(int, int);
	};

CMatrix::CMatrix(int paramNumRows, int paramNumCols)
	{
	CMatrixNode *rowptr;
	CMatrixNode *colptr;
	int row, col;

	numrows = paramNumRows;
	numcols = paramNumCols;

	head = new CMatrixNode;
	rowptr = head;

	for(row = 0; row < numrows; row++)
		{
		colptr = rowptr;

		for(col = 0; col < numcols-1; col++)
			{
			colptr->right = new CMatrixNode;
			colptr->right->left = colptr;
			colptr = colptr->right;
			}

		if (row < numrows -1)
			{
			rowptr->lower = new CMatrixNode;
			rowptr->lower->upper = rowptr;
			rowptr = rowptr->lower;
			}
		}
	}

CMatrix::~CMatrix()
	{
	CMatrixNode *ptr;

	ptr = head;

	while(ptr->lower != NULL)
		ptr = ptr->lower;
	while(ptr->right != NULL)
		ptr = ptr->right;

	while(ptr->upper != NULL)
		{
		while(ptr->left != NULL)
			{
			ptr = ptr->left;
			delete ptr->right;
			}
		ptr = ptr->upper;
		delete ptr->lower;
		}

	while(ptr->right != NULL)
		ptr = ptr->right;
	while(ptr->left != NULL)
		{
		ptr = ptr->left;
		delete ptr->right;
		}

	delete head;
	}

CMatrix::Set(int paramRow, int paramCol, int paramVal)
	{
	CMatrixNode *ptr;
	int row, col;

	ptr = head;

	for(row = 0; row < paramRow; row++)
		ptr = ptr->lower;

	for(col = 0; col < paramCol; col++)
		ptr = ptr->right;

	ptr->value = paramVal;
	}

int CMatrix::Get(int paramRow, int paramCol)
	{
	CMatrixNode *ptr;
	int row, col;

	ptr = head;

	for(row = 0; row < paramRow; row++)
		ptr = ptr->lower;

	for(col = 0; col < paramCol; col++)
		ptr = ptr->right;

	return ptr->value;
	}

///////////////////////////////////////////////////////////////////////////////////////////////////

class CMesh
	{
	public:

	// unaltered data loaded from .3do file
	char   jkName[_MAX_FNAME];
	double jkRadius;
	int    jkGeometryMode;
	int    jkLightingMode;
	int    jkTextureMode;

	int numVerts;
	JK_vertex *jkVerts;

	int numTxverts;
	JK_txvert *jkTxverts;

	JK_normal *jkVertnormals;

	int numjkFaces;
	JK_face *jkFaces;

	JK_normal *jkFacenormals;

	// ForceRemap program-specific data
	JK_offset jkOffset;

	GL_vertex *glVerts;
	GL_txvert *glTxverts;

	int numglGroups;
	GL_group *glGroups;

	GL_rendermode glRenderMode;

	// output information for saving
	int numOutputTxverts;
	JK_txvert *outputTxverts;

	JK_face *outputFaces;

	// ForceRemap program-specific functions
	void Deselect_AllGroups();
	void Select_Group(int);
	};

void CMesh::Deselect_AllGroups()
	{
	for(int g = 0; g < numglGroups; g++)
		glGroups[g].is_selected = false;
	}

void CMesh::Select_Group(int groupID)
	{
	glGroups[groupID].is_selected = true;
	}

class CHNode
	{
	public:

	char flags[64];
	char type [64];

	int mesh;
	int parent;
	int child;
	int sibling;
	int numChildren;

	// default offset of this node from parent.
	// (used in-game while not being dictated by keyframe animation)
	double x, y, z;

	// default rotation of this node.
	// (used in-game while not being dictated by keyframe animation)
	double pitch, yaw, roll;

	// permanent offset of this node from parent.
	// (effective always in addition to keyframe animation)
	double pivotX, pivotY, pivotZ;

	char hNodeName[_MAX_FNAME];
	};

///////////////////////////////////////////////////////////////////////////////////////////////////

class CCmpFileHeader
	{
	public:

	char cmpID[4];
	int  version;
	int  transparency;
	unsigned char padding[52];
	};

class CCmpPalette
	{
	public:

	unsigned char palette[256*3];
	};

void LoadDefaultCmpFile()
	{
	const unsigned char dfltCmp[] = {
		0,0,0,       0,255,0,     0,203,0,     0,155,0,     0,107,0,     0,59,0,      255,0,0,     203,0,0,     155,0,0,     107,0,0,     59,0,0,      247,255,0,   215,163,0,   175,87,0,    135,31,0,    95,0,0,
		255,255,255, 223,231,255, 195,215,255, 163,195,255, 135,175,255, 255,171,0,   255,159,0,   255,147,0,   255,131,0,   255,111,0,   255,91,0,    255,71,0,    255,51,0,    255,35,0,    255,15,0,    0,0,255,
		253,253,253, 247,247,247, 239,239,239, 227,227,227, 219,219,219, 211,211,211, 203,203,203, 195,195,195, 187,187,187, 179,179,179, 171,171,171, 163,163,163, 155,155,155, 147,147,147, 139,139,139, 131,131,131,
		123,123,123, 115,115,115, 107,107,107, 99,99,99,    87,87,87,    79,79,79,    71,71,71,    63,63,63,    55,55,55,    47,47,47,    39,39,39,    31,31,31,    23,23,23,    15,15,15,    7,7,7,       0,0,0,
		191,199,223, 183,191,215, 179,183,207, 171,179,203, 163,171,195, 159,163,187, 151,159,183, 147,151,175, 139,143,167, 135,139,163, 127,131,155, 119,123,147, 115,119,139, 107,111,135, 103,107,127, 95,99,119,
		91,95,115,   87,87,107,   79,83,99,    75,75,95,    67,71,87,    63,63,79,    55,59,75,    51,51,67,    47,47,59,    39,43,55,    35,35,47,    31,31,39,    23,27,35,    19,19,27,    15,15,19,    11,11,15,
		255,207,179, 231,175,143, 207,143,111, 183,119,87,  159,91,63,   135,71,43,   111,51,27,   87,35,15,    255,255,0,   227,195,0,   199,143,0,   171,99,0,    147,63,0,    119,31,0,    91,11,0,     67,0,0,
		223,255,167, 207,239,135, 191,223,103, 179,207,75,  167,191,51,  159,175,31,  151,159,11,  143,147,0,   199,99,31,   183,87,23,   171,75,19,   155,63,11,   143,55,7,    127,47,7,    115,39,0,    103,31,0,
		251,0,0,     227,0,0,     199,0,0,     171,0,0,     143,0,0,     115,0,0,     87,0,0,      57,0,0,      127,163,199, 95,127,171,  67,95,147,   43,67,123,   23,39,95,    11,19,71,    0,7,47,      0,0,23,
		195,115,71,  183,107,63,  175,99,59,   163,91,51,   155,87,47,   147,79,43,   135,71,35,   127,67,31,   115,59,27,   107,55,23,   99,47,19,    87,43,15,    79,39,15,    67,31,11,    59,27,7,     51,23,7,
		255,231,179, 239,211,155, 223,195,135, 211,179,119, 195,163,99,  183,147,83,  167,135,71,  151,119,55,  139,103,43,  123,91,31,   111,79,23,   95,67,15,    79,55,11,    67,43,7,     51,31,0,     39,23,0,
		131,231,103, 115,207,83,  99,183,67,   83,159,55,   71,139,43,   59,115,31,   47,91,23,    35,71,15,    255,167,255, 223,127,231, 195,95,207,  163,67,183,  135,43,159,  107,23,135,  79,7,111,    55,0,91,
		255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,
		255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,
		255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,
		255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,0,255,   255,255,255 };

	for (int i = 0; i < 256*3; i++)
		gPalette->palette[i] = dfltCmp[i];

	gCmpHasBeenLoaded = true;
	}

bool LoadCmpFile()
	{
	if (gPalette != NULL)
		delete gPalette;
	gPalette = new CCmpPalette;

	if (!strcmpi((char*)gCmpPath, (char*)gszDfltCmp))
		LoadDefaultCmpFile();

	else
		{
		FILE* filePtr;
		CCmpFileHeader fileheader;

		if ((filePtr = fopen((char*)gCmpPath, "rb")) == NULL)
			{
			if (!gCmpHasBeenLoaded)
				LoadDefaultCmpFile();

			Msg(ghwndMain, "Invalid CMP File", (char*)gCmpPath);

			strcpy((char*)gCmpPath, (char*)gszDfltCmp);
			}
		else
			{
			fread(&fileheader, sizeof(CCmpFileHeader), 1, filePtr);

			if ((fileheader.cmpID[0] == 'C') && (fileheader.cmpID[1] == 'M') && (fileheader.cmpID[2] == 'P'))
				{
				fread(gPalette, sizeof(CCmpPalette), 1, filePtr);
				gCmpHasBeenLoaded = true;
				}
			else
				{
				if (!gCmpHasBeenLoaded)
					LoadDefaultCmpFile();

				Msg(ghwndMain, "Invalid CMP File", (char*)gCmpPath);

				strcpy((char*)gCmpPath, (char*)gszDfltCmp);
				}

			fclose(filePtr);
			}
		}

	return true;
	}

class CMatFileHeader
	{
	public:

	char matID[4];       // = "MAT " (notice space after MAT)
	int  version;        // = 0x32 ('2')
	int  type;           // 0 = color, 1 = ?, 2 = texture
	int  numTextures0;   // number of textures or colors
	int  numTextures1;   // for color MATs = 0, for texture MATs = numTextures
	int  unknown;        //
	int  bits;           // = 8, 16
	int  padding[12];    // some padding?
	};

class CMatColorHeader
	{
	public:

	int type;            // 0 = color, 8 = texture
	int colorNum;        // color index from the CMP palette
	int array[4];        // each = 0x3F800000 (check cmp header)
	};

class CMatTextureHeader
	{
	public:

	int type;            // 0 = color, 8 = texture
	int colorNum;        // unknown use (background color?)
	int int0;            // each = 0x3F800000 (check cmp header)
	int int1;
	int int2;
	int int3;
	int unknown0;        // unknown
	int unknown1;
	int longint;         // = 0xBFF78482
	int currentTXnum;    // number of corresponding texture, 0..NumTextures-1
	};

class CMatDataHeader
	{
	public:

	int width;           // width of first MipMap, must be divisible by 2
	int height;          // height of first MipMap, must be divisible by 2
	int padding0;        // padding = 0
	int padding1;
	int padding2;
	int numMipmaps;      // number of mipmaps in texture (largest one first)
	};

class CMaterial
	{
	public:

	// filenames
	char genericname[_MAX_FNAME]; // "filename"
	char pathname   [_MAX_PATH ]; // "x:\...\filename.mat" or "x:\...\filename.bmp"

	// texture data
	bool file_error;
	int width, height;
	unsigned char* textureData; // raw RGB data
	unsigned int glTexture[1]; // actual OpenGL texture object
	bool is8bit;
	bool user_specified;

	// functions
	Create_glTexture(char*);
	void Load_MatFile();
	void Load_BmpFile();
	void Reload_Texture();
	};

void CMaterial::Reload_Texture()
	{
	Create_glTexture(pathname);
	}

CMaterial::Create_glTexture(char* newpathname)
	{
	FILE* fileptr;

	if (!strcmp(newpathname, gszEmptyString))
		{
		user_specified = false;

		if (gBmpFirst)
			{
			strcpy(pathname, gszEmptyString);
			strcpy(pathname, gszProjectFolder);
			strcat(pathname, "bmp\\");
			strcat(pathname, genericname);
			strcat(pathname, ".BMP");

			if ((fileptr = fopen(pathname, "rb")) == NULL)
				{
				strcpy(pathname, gszEmptyString);
				strcpy(pathname, gszProjectFolder);
				strcat(pathname, genericname);
				strcat(pathname, ".BMP");

				if (((fileptr = fopen(pathname, "rb")) == NULL) && (gMatSecond))
					{
					strcpy(pathname, gszEmptyString);
					strcpy(pathname, gszProjectFolder);
					strcat(pathname, "mat\\");
					strcat(pathname, genericname);
					strcat(pathname, ".MAT");

					if ((fileptr = fopen(pathname, "rb")) == NULL)
						{
						strcpy(pathname, gszEmptyString);
						strcpy(pathname, gszProjectFolder);
						strcat(pathname, genericname);
						strcat(pathname, ".MAT");
						}
					}
				}
			}
		else
			{
			strcpy(pathname, gszEmptyString);
			strcpy(pathname, gszProjectFolder);
			strcat(pathname, "mat\\");
			strcat(pathname, genericname);
			strcat(pathname, ".MAT");

			if ((fileptr = fopen(pathname, "rb")) == NULL)
				{
				strcpy(pathname, gszEmptyString);
				strcpy(pathname, gszProjectFolder);
				strcat(pathname, genericname);
				strcat(pathname, ".MAT");

				if (((fileptr = fopen(pathname, "rb")) == NULL) && (gBmpSecond))
					{
					strcpy(pathname, gszEmptyString);
					strcpy(pathname, gszProjectFolder);
					strcat(pathname, "bmp\\");
					strcat(pathname, genericname);
					strcat(pathname, ".BMP");

					if ((fileptr = fopen(pathname, "rb")) == NULL)
						{
						strcpy(pathname, gszEmptyString);
						strcpy(pathname, gszProjectFolder);
						strcat(pathname, genericname);
						strcat(pathname, ".BMP");
						}
					}
				}
			}
		}

	else
		{
		user_specified = true;

		strcpy(pathname, newpathname);

		int pathpos = strlen(pathname);
		while(pathname[--pathpos] != '\\');
		pathpos++;
		int genericpos = 0;
		while(pathname[pathpos] != '.')
			genericname[genericpos++] = pathname[pathpos++];
		genericname[genericpos] = '\0';
		}

	char ext = pathname[strlen(pathname)-1];

	// initialize the material's flags.
	file_error = true;
	textureData = NULL;
	width = height = 0;
	is8bit = false;

	// load the texture data.
	if ((ext == 'P') || (ext == 'p'))
		Load_BmpFile();
	else
		Load_MatFile();

	// generate the new OpenGL texture object and replace the old one (if exists).
	glDeleteTextures (1, glTexture);
	glGenTextures    (1, glTexture);
	glBindTexture    (GL_TEXTURE_2D, glTexture[0]);
	glTexParameteri  (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri  (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	if ((ext == 'P') || (ext == 'p'))
		glTexImage2D     (GL_TEXTURE_2D, 0, 3, width, height, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, textureData);
	else
		glTexImage2D     (GL_TEXTURE_2D, 0, 3, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, textureData);

	// the raw texture data is no longer needed.
	free(textureData);
	}

void CMaterial::Load_MatFile()
	{
	FILE*             filePtr;
	CMatFileHeader    matFileHeader;
	CMatColorHeader   matColorHeader;
	CMatTextureHeader matTextureHeader;
	CMatDataHeader    matDataHeader;

	int i;
	int size;
	unsigned char* tempData;

	if ((filePtr = fopen(pathname, "rb")) == NULL)
		{
		if (!user_specified && !gBmpFirst && gBmpSecond)
			{
			pathname[strlen(pathname)-3] = 'B';
			pathname[strlen(pathname)-2] = 'M';
			pathname[strlen(pathname)-1] = 'P';
			Load_BmpFile();
			}
		}
	else
		{
		fread(&matFileHeader, sizeof(CMatFileHeader), 1, filePtr);

		if (matFileHeader.type == 0)
			{
			is8bit = true;

			fread(&matColorHeader, sizeof(CMatColorHeader), 1, filePtr);
			width = height = 8;
			size = width * height;
			textureData = (unsigned char*)malloc(3*size);

			for(i = 0; i < size; i++)
				{
				textureData[3*i  ] = gPalette->palette[3*matColorHeader.colorNum   ];
				textureData[3*i+1] = gPalette->palette[3*matColorHeader.colorNum +1];
				textureData[3*i+2] = gPalette->palette[3*matColorHeader.colorNum +2];
				}
			}
		else
			{
			fread(&matTextureHeader, sizeof(CMatTextureHeader), 1, filePtr);
			fread(&matDataHeader, sizeof(CMatDataHeader), 1, filePtr);

			width  = matDataHeader.width;
			height = matDataHeader.height;

			size = width*height;
			tempData = (unsigned char*)malloc(3*size);
			textureData = (unsigned char*)malloc(3*size);

			if(matFileHeader.bits == 8)
				{
				is8bit = true;

				unsigned char* colors = (unsigned char *)malloc(size);
				for(i = 0; i < size; i++)
					fread(colors+i, 1, 1, filePtr);

				for(i = 0; i < size; i++)
					{
					tempData[3*i  ] = gPalette->palette[3*colors[i]  ];
					tempData[3*i+1] = gPalette->palette[3*colors[i]+1];
					tempData[3*i+2] = gPalette->palette[3*colors[i]+2];
					}
				}
			else
				{
				unsigned char* rgxb5515 = (unsigned char*)malloc(2);

				for(i = 0; i < 3*size; i += 3)
					{
					fread(rgxb5515+1, 1, 1, filePtr);
					fread(rgxb5515,   1, 1, filePtr);

					int r = i;
					int g = i+1;
					int b = i+2;

					tempData[r] = rgxb5515[0] & 0xF8;
					tempData[g] = ((rgxb5515[0] << 5) | (rgxb5515[1] >> 3)) & 0xF8;
					tempData[b] = rgxb5515[1] << 3;

					tempData[r] = int(tempData[r] * RGB_16_TO_24);
					tempData[g] = int(tempData[g] * RGB_16_TO_24);
					tempData[b] = int(tempData[b] * RGB_16_TO_24);
					}

				free(rgxb5515);
				}

			// flip the rows
			for(int indent = 0; indent < 3*size; indent += 3*width)
				for(int col = 0; col < 3*width; col ++)
					textureData[3*size-3*width-indent+col] = tempData[indent+col];
			}

		file_error = false;
		fclose(filePtr);
		}
	}

void CMaterial::Load_BmpFile()
	{
	FILE*             filePtr;
	BITMAPFILEHEADER  bitmapFileHeader;
	BITMAPINFOHEADER  bmpInfoHeader;

	if ((filePtr = fopen(pathname, "rb")) == NULL)
		{
		if (!user_specified && !gMatFirst && gMatSecond)
			{
			pathname[strlen(pathname)-3] = 'M';
			pathname[strlen(pathname)-2] = 'A';
			pathname[strlen(pathname)-1] = 'T';
			Load_MatFile();
			}
		}
	else
		{
		fread(&bitmapFileHeader, sizeof(BITMAPFILEHEADER), 1, filePtr);
		fread(&bmpInfoHeader, sizeof(BITMAPINFOHEADER), 1, filePtr);

		width  = bmpInfoHeader.biWidth;
		height = bmpInfoHeader.biHeight;

		textureData = (unsigned char*)malloc(bmpInfoHeader.biSizeImage);

		fseek(filePtr, bitmapFileHeader.bfOffBits, SEEK_SET);
		fread(textureData, 1, bmpInfoHeader.biSizeImage, filePtr);

		fclose(filePtr);

		file_error = false;
		}
	}

class C3DOModel
	{
	public:

		ifstream loadfile;

		double jk3doVersion;

		int numMaterials;
		CMaterial *materials;

		double jkModelRadius;
		JK_offset jkInsertOffset;

		int jkNumGeosets;

		int numMeshes;
		CMesh *meshes;

		int numHNodes;
		CHNode *hNodes;

	double glMinX, glMaxX, glMinY, glMaxY, glMinZ, glMaxZ;
	GL_offset glViewOffset;
	double radius;

	C3DOModel();
	~C3DOModel();

	bool Load_File(char *);
	bool FindNextToken();
	bool FindKeyword(char *);
	bool GetInt(int *);
	bool GetDouble(double *);
	bool GetString(char *);

	bool Initialize_Model();
	void Break_Txverts(int);

	bool Save_File(char *);
	void Prepare_Model();

	void Display();

	void Deselect_AllMeshes();
	void Change_Tex(int, int, char *, char *);

	void Reload8bitTextures();
	};

void C3DOModel::Reload8bitTextures(void)
	{
	for(int m = 0; m < numMaterials; m++)
		{
		if (materials[m].is8bit)
			{
			materials[m].Reload_Texture();
			}
		}
	}

C3DOModel::C3DOModel()
	{
	materials = NULL;

	meshes = NULL;

	hNodes = NULL;
	}

C3DOModel::~C3DOModel()
	{
	int m;

	if (materials != NULL)
		free(materials);

	if (meshes != NULL)
		{
		for(m = 0; m < numMeshes; m++)
			{
			free(meshes[m].jkVerts);
			free(meshes[m].jkTxverts);
			free(meshes[m].jkVertnormals);
			free(meshes[m].jkFaces);
			free(meshes[m].jkFacenormals);

			free(meshes[m].glVerts);
			free(meshes[m].glTxverts);
			free(meshes[m].glGroups);

			free(meshes[m].outputTxverts);
			free(meshes[m].outputFaces);
			}

		free(meshes);
		}

	if (hNodes != NULL)
		free(hNodes);
	}

bool C3DOModel::Load_File(char *szPathname)
	{
	int m, v, f, p, h;

	char trash[_MAX_FNAME];

	// open the 3do file
	loadfile.open(szPathname, ios::nocreate, filebuf::sh_none);

	if (loadfile.fail())
		return false;

	// load HEADER data
	if (!FindKeyword("SECTION:"))                                     return false;
	if (!FindKeyword("HEADER"))                                       return false;

	if (!FindKeyword("3DO"))                                          return false;
	if (!GetDouble(&jk3doVersion))                                    return false;

	// load MODELRESOURCE data
	if (!FindKeyword("SECTION:"))                                     return false;
	if (!FindKeyword("MODELRESOURCE"))                                return false;

	if (!FindKeyword("MATERIALS"))                                    return false;
	if (!GetInt(&numMaterials))                                       return false;

	materials = (CMaterial*)malloc(sizeof(CMaterial)*numMaterials);
	if (materials == NULL) { loadfile.close(); return false; }

	for(m = 0; m < numMaterials; m++)
		{
		if (!GetString(trash))                                         return false;
		if (!GetString(materials[m].genericname))                     return false;
		materials[m].genericname[strlen(materials[m].genericname)-4] = '\0';
		}

	// load GEOMETRYDEF data
	if (!FindKeyword("SECTION:"))                                     return false;
	if (!FindKeyword("GEOMETRYDEF"))                                  return false;

	if (!FindKeyword("RADIUS"))                                       return false;
	if (!GetDouble(&jkModelRadius))                                   return false;

	if (!FindKeyword("INSERT"))                                       return false;
	if (!FindKeyword("OFFSET"))                                       return false;
	if (!GetDouble(&jkInsertOffset.x))                                return false;
	if (!GetDouble(&jkInsertOffset.y))                                return false;
	if (!GetDouble(&jkInsertOffset.z))                                return false;

	if (!FindKeyword("GEOSETS"))                                      return false;
	if (!GetInt(&jkNumGeosets))                                       return false;

	if (!FindKeyword("MESHES"))                                       return false;
	if (!GetInt(&numMeshes))                                          return false;
	if (numMeshes < 1)
		{ loadfile.close(); return false; }

	// load meshes
	meshes = (CMesh*)malloc(sizeof(CMesh)*numMeshes);
	if (meshes == NULL) { loadfile.close(); return false; }

	for(m = 0; m < numMeshes; m++)
		{
		meshes[m].jkVerts = NULL;
		meshes[m].jkTxverts = NULL;
		meshes[m].jkVertnormals = NULL;
		meshes[m].jkFaces = NULL;
		meshes[m].jkFacenormals = NULL;

		meshes[m].glVerts = NULL;
		meshes[m].glTxverts = NULL;
		meshes[m].glGroups = NULL;

		meshes[m].outputTxverts = NULL;
		meshes[m].outputFaces = NULL;
		}

	for(m = 0; m < numMeshes; m++)
		{
		if (!FindKeyword("MESH"))                                      return false;

		if (!FindKeyword("NAME"))                                      return false;
		if (!GetString(meshes[m].jkName))                              return false;

		if (!FindKeyword("RADIUS"))                                    return false;
		if (!GetDouble(&meshes[m].jkRadius))                           return false;

		if (!FindKeyword("GEOMETRYMODE"))                              return false;
		if (!GetInt(&meshes[m].jkGeometryMode))                        return false;

		if (!FindKeyword("LIGHTINGMODE"))                              return false;
		if (!GetInt(&meshes[m].jkLightingMode))                        return false;

		if (!FindKeyword("TEXTUREMODE"))                               return false;
		if (!GetInt(&meshes[m].jkTextureMode))                         return false;

		// load vertices
		if (!FindKeyword("VERTICES"))                                  return false;
		if (!GetInt(&meshes[m].numVerts))                              return false;

		meshes[m].jkVerts = (JK_vertex*)malloc(sizeof(JK_vertex)*meshes[m].numVerts);
		if (meshes[m].jkVerts == NULL) { loadfile.close(); return false; }

		for(v = 0; v < meshes[m].numVerts; v++)
			{
			if (!GetString(trash))                                      return false;
			if (!GetDouble(&meshes[m].jkVerts[v].x))                    return false;
			if (!GetDouble(&meshes[m].jkVerts[v].y))                    return false;
			if (!GetDouble(&meshes[m].jkVerts[v].z))                    return false;
			if (!GetDouble(&meshes[m].jkVerts[v].i))                    return false;
			}

		// load texture vertices
		if (!FindKeyword("TEXTURE"))                                   return false;
		if (!FindKeyword("VERTICES"))                                  return false;
		if (!GetInt(&meshes[m].numTxverts))                            return false;

		meshes[m].jkTxverts = (JK_txvert*)malloc(sizeof(JK_txvert)*meshes[m].numTxverts);
		if (meshes[m].jkTxverts == NULL) { loadfile.close(); return false; }

		for(v = 0; v < meshes[m].numTxverts; v++)
			{
			if (!GetString(trash))                                      return false;
			if (!GetDouble(&meshes[m].jkTxverts[v].s))                  return false;
			if (!GetDouble(&meshes[m].jkTxverts[v].t))                  return false;
			}

		// load vertex normals
		if (!FindKeyword("VERTEX"))                                    return false;
		if (!FindKeyword("NORMALS"))                                   return false;

		meshes[m].jkVertnormals = (JK_normal*)malloc(sizeof(JK_normal)*meshes[m].numVerts);
		if (meshes[m].jkVertnormals == NULL) { loadfile.close(); return false; }

		for(v = 0; v < meshes[m].numVerts; v++)
			{
			if (!GetString(trash))                                      return false;
			if (!GetDouble(&meshes[m].jkVertnormals[v].x))              return false;
			if (!GetDouble(&meshes[m].jkVertnormals[v].y))              return false;
			if (!GetDouble(&meshes[m].jkVertnormals[v].z))              return false;
			}

		// load faces
		if (!FindKeyword("FACES"))                                     return false;
		if (!GetInt(&meshes[m].numjkFaces))                            return false;

		meshes[m].jkFaces = (JK_face*)malloc(sizeof(JK_face)*meshes[m].numjkFaces);
		if (meshes[m].jkFaces == NULL) { loadfile.close(); return false; }

		for(f = 0; f < meshes[m].numjkFaces; f++)
			{
			if (!GetString(trash))                                      return false;
			if (!GetInt(&meshes[m].jkFaces[f].matID))                   return false;
			if (!GetString(meshes[m].jkFaces[f].jkType))                return false;
			if (!GetInt(&meshes[m].jkFaces[f].jkGeo))                   return false;
			if (!GetInt(&meshes[m].jkFaces[f].jkLight))                 return false;
			if (!GetInt(&meshes[m].jkFaces[f].jkTex))                   return false;
			if (!GetDouble(&meshes[m].jkFaces[f].jkExtralight))         return false;
			if (!GetInt(&meshes[m].jkFaces[f].numjkPairs))              return false;

			// load pairs
			meshes[m].jkFaces[f].jkPairs = (JK_pair*)malloc(sizeof(JK_pair)*meshes[m].jkFaces[f].numjkPairs);
			if (meshes[m].jkFaces[f].jkPairs == NULL) { loadfile.close(); return false; }

			for(p = 0; p < meshes[m].jkFaces[f].numjkPairs; p++)
				{
				if (!GetInt(&meshes[m].jkFaces[f].jkPairs[p].vertexID))  return false;
				if (!GetInt(&meshes[m].jkFaces[f].jkPairs[p].txvertID))  return false;
				}
			} // end face loop

		// load face normals
		if (!FindKeyword("FACE"))                                      return false;
		if (!FindKeyword("NORMALS"))                                   return false;

		meshes[m].jkFacenormals = (JK_normal*)malloc(sizeof(JK_normal)*meshes[m].numjkFaces);
		if (meshes[m].jkFacenormals == NULL) { loadfile.close(); return false; }

		for(f = 0; f < meshes[m].numjkFaces; f++)
			{
			if (!GetString(trash))                                      return false;
			if (!GetDouble(&meshes[m].jkFacenormals[f].x))              return false;
			if (!GetDouble(&meshes[m].jkFacenormals[f].y))              return false;
			if (!GetDouble(&meshes[m].jkFacenormals[f].z))              return false;
			}
		} // end mesh loop

	// load the HIERARCHYDEF data
	if (!FindKeyword("SECTION:"))                                     return false;
	if (!FindKeyword("HIERARCHYDEF"))                                 return false;

	if (!FindKeyword("HIERARCHY"))                                    return false;
	if (!FindKeyword("NODES"))                                        return false;
	if (!GetInt(&numHNodes))                                          return false;

	hNodes = (CHNode*)malloc(sizeof(CHNode)*numHNodes);
	if (hNodes == NULL) { loadfile.close(); return false; }

	for(h = 0; h < numHNodes; h++)
		{
		if (!GetString(trash))                                         return false;
		if (!GetString(hNodes[h].flags))                               return false;
		if (!GetString(hNodes[h].type))                                return false;
		if (!GetInt(&hNodes[h].mesh))                                  return false;
		if (!GetInt(&hNodes[h].parent))                                return false;
		if (!GetInt(&hNodes[h].child))                                 return false;
		if (!GetInt(&hNodes[h].sibling))                               return false;
		if (!GetInt(&hNodes[h].numChildren))                           return false;
		if (!GetDouble(&hNodes[h].x))                                  return false;
		if (!GetDouble(&hNodes[h].y))                                  return false;
		if (!GetDouble(&hNodes[h].z))                                  return false;
		if (!GetDouble(&hNodes[h].pitch))                              return false;
		if (!GetDouble(&hNodes[h].yaw))                                return false;
		if (!GetDouble(&hNodes[h].roll))                               return false;
		if (!GetDouble(&hNodes[h].pivotX))                             return false;
		if (!GetDouble(&hNodes[h].pivotY))                             return false;
		if (!GetDouble(&hNodes[h].pivotZ))                             return false;
		if (!GetString(hNodes[h].hNodeName))                           return false;
		}

	loadfile.close();

	return true;
	}

bool C3DOModel::FindNextToken()
	{
	char next;
	char trash[1];
	char token[_MAX_FNAME];

	// if end of file, return fail
	if (loadfile.eof())
		{
		loadfile.close();
		return false;
		}

	// eat whitespace and commas
	loadfile.eatwhite();
	next = loadfile.peek();
	if (next == ',')
		{
		loadfile.read(trash, 1);
		next = loadfile.peek();
		}

	// if end of file, return fail
	if (next == EOF)
		{
		loadfile.close();
		return false;
		}

	// if comment, skip that line and try again
	else if (next == '#')
		{
		loadfile.getline(token, _MAX_FNAME, '\n');
		if (!FindNextToken())
			return false;
		}

	return true;
	}

bool C3DOModel::FindKeyword(char *keyword)
	{
	char token[_MAX_FNAME];

	if (!FindNextToken())
		return false;

	loadfile >> token;
	while (strcmp(token, keyword))
		{
		if (!FindNextToken())
			return false;

		loadfile >> token;
		}

	if (!FindNextToken())
		return false;

	return true;
	}

bool C3DOModel::GetInt(int *data)
	{
	if (!FindNextToken())
		return false;

	loadfile >> *data;

	return true;
	}

bool C3DOModel::GetDouble(double *data)
	{
	if (!FindNextToken())
		return false;

	loadfile >> *data;

	return true;
	}

bool C3DOModel::GetString(char *data)
	{
	if (!FindNextToken())
		return false;

	loadfile >> data;

	return true;
	}

bool C3DOModel::Initialize_Model()
	{
	int m, h, v, g, f, p, j;

	CHNode *parentnode;

	int groupArray[256][2];
	bool matched;

	double maxS, maxT;

	char warning[512];

	// initialize dimensions to opposite extreme values
	glMinX = glMinY = glMinZ =  1.7e300;
	glMaxX = glMaxY = glMaxZ = -1.7e300;
	radius = -1.7e300;

	// load bitmap files and generate OpenGL textures
	for(m = 0; m < numMaterials; m++)
		materials[m].Create_glTexture(gszEmptyString);

	for(h = 0; h < numHNodes; h++)
		{
/*
		// set special offset for K_RHAND
		if (!(strcmpi(hNodes[h].hNodeName, "K_RHAND")))
			{
			hNodes[h].pivotX = -0.002488;
			hNodes[h].pivotY =  0.009120;
			hNodes[h].pivotZ = -0.005542;
			}
*/
		// calculate total hierarchical offset for each node
		if (hNodes[h].mesh >= 0)
			{
			meshes[hNodes[h].mesh].jkOffset.x = hNodes[h].x + hNodes[h].pivotX;
			meshes[hNodes[h].mesh].jkOffset.y = hNodes[h].y + hNodes[h].pivotY;
			meshes[hNodes[h].mesh].jkOffset.z = hNodes[h].z + hNodes[h].pivotZ;

			parentnode = &hNodes[hNodes[h].parent];

			while (parentnode->parent >= 0)
				{
				meshes[hNodes[h].mesh].jkOffset.x += parentnode->x;
				meshes[hNodes[h].mesh].jkOffset.y += parentnode->y;
				meshes[hNodes[h].mesh].jkOffset.z += parentnode->z;

				parentnode = &hNodes[parentnode->parent];
				}
			}
		}

	for(m = 0; m < numMeshes; m++)
		{
		// generate GL vertices for the mesh using hierarchical offsets
		meshes[m].glVerts = (GL_vertex*)malloc(sizeof(GL_vertex)*meshes[m].numVerts);
		if (meshes[m].glVerts == NULL)
			{
			MessageBox(ghwndMain, "Could not allocate memory for: glVerts.", "Error", MB_APPLMODAL);
			return false;
			}
		for(v = 0; v < meshes[m].numVerts; v++)
			{
			meshes[m].glVerts[v].x = meshes[m].jkVerts[v].x + meshes[m].jkOffset.x;
			meshes[m].glVerts[v].y = meshes[m].jkVerts[v].y + meshes[m].jkOffset.y;
			meshes[m].glVerts[v].z = meshes[m].jkVerts[v].z + meshes[m].jkOffset.z;

			// also update the min and max dimensions of the model
			if (meshes[m].glVerts[v].x < glMinX) glMinX = meshes[m].glVerts[v].x;
			else if (meshes[m].glVerts[v].x > glMaxX) glMaxX = meshes[m].glVerts[v].x;

			if (meshes[m].glVerts[v].y < glMinY) glMinY = meshes[m].glVerts[v].y;
			else if (meshes[m].glVerts[v].y > glMaxY) glMaxY = meshes[m].glVerts[v].y;

			if (meshes[m].glVerts[v].z < glMinZ) glMinZ = meshes[m].glVerts[v].z;
			else if (meshes[m].glVerts[v].z > glMaxZ) glMaxZ = meshes[m].glVerts[v].z;
			}

		// generate glTxverts for the mesh from the jkTxverts
		meshes[m].glTxverts = (GL_txvert*)malloc(sizeof(GL_txvert)*meshes[m].numTxverts);
		if (meshes[m].glTxverts == NULL)
			{
			MessageBox(ghwndMain, "Could not allocate memory for: glTxverts.", "Error", MB_APPLMODAL);
			return false;
			}
		for(j = 0; j < meshes[m].numjkFaces; j++)
			{
			for(p = 0; p < meshes[m].jkFaces[j].numjkPairs; p++)
				{
				if (false == materials[meshes[m].jkFaces[j].matID].file_error)
					{
					meshes[m].glTxverts[meshes[m].jkFaces[j].jkPairs[p].txvertID].s =   meshes[m].jkTxverts[meshes[m].jkFaces[j].jkPairs[p].txvertID].s / materials[meshes[m].jkFaces[j].matID].width;
					meshes[m].glTxverts[meshes[m].jkFaces[j].jkPairs[p].txvertID].t = - meshes[m].jkTxverts[meshes[m].jkFaces[j].jkPairs[p].txvertID].t / materials[meshes[m].jkFaces[j].matID].height;
					}
				}
			}

		// determine the number of glGroups and set the groupID for each jkFace
		// also store the number of faces belonging to each group
		for(g = 0; g < 256; g++)
			groupArray[g][0] = groupArray[g][1] = 0;

		meshes[m].numglGroups = 0;
		if (meshes[m].numjkFaces > 0)
			{
			groupArray[0][0] = meshes[m].jkFaces[0].matID;
			groupArray[0][1] = 1;

			meshes[m].jkFaces[0].groupID = 0;
			meshes[m].numglGroups = 1;

			for(f = 1; f < meshes[m].numjkFaces; f++)
				{
				matched = false;

				for(g = 0; g < meshes[m].numglGroups; g++)
					{
					if (meshes[m].jkFaces[f].matID == groupArray[g][0])
						{
						matched = true;

						groupArray[g][1]++;
						meshes[m].jkFaces[f].groupID = g;

						g = meshes[m].numglGroups;
						}

					}
				if (!matched)
					{
					groupArray[meshes[m].numglGroups][0] = meshes[m].jkFaces[f].matID;
					groupArray[meshes[m].numglGroups][1] = 1;

					meshes[m].jkFaces[f].groupID = meshes[m].numglGroups;
					meshes[m].numglGroups++;
					}
				}
			}

		// make the groups have a unique set of txverts
		Break_Txverts(m);

		// create the glGroups and set up all the data they contain
		meshes[m].glGroups = (GL_group*)malloc(sizeof(GL_group)*meshes[m].numglGroups);
		if (meshes[m].glGroups == NULL)
			{
			MessageBox(ghwndMain, "Could not allocate memory for: glGroups.", "Error", MB_APPLMODAL);
			return false;
			}
		for(g = 0; g < meshes[m].numglGroups; g++)
			{
			// set the original flag telling the renderer to access the Mesh's glTxverts and not the glGroup's
			meshes[m].glGroups[g].is_original = true;
			meshes[m].glGroups[g].is_selected = false;

			strcpy(meshes[m].glGroups[g].glGroupName, materials[groupArray[g][0]].genericname);
			meshes[m].glGroups[g].matID = groupArray[g][0];
			meshes[m].glGroups[g].numglFaces = groupArray[g][1];

			meshes[m].glGroups[g].glTxverts = NULL;

			meshes[m].glGroups[g].glFaces = (GL_face*)malloc(sizeof(GL_face)*meshes[m].glGroups[g].numglFaces);
			if (meshes[m].glGroups[g].glFaces == NULL)
				{
				MessageBox(ghwndMain, "Could not allocate memory for: glFaces.", "Error", MB_APPLMODAL);
				return false;
				}

			f = 0;
			for(j = 0; j < meshes[m].numjkFaces; j++)
				{
				if (meshes[m].jkFaces[j].groupID == g)
					{
					meshes[m].glGroups[g].glFaces[f].jkFaceID = j;

					meshes[m].glGroups[g].glFaces[f].numglPairs = meshes[m].jkFaces[j].numjkPairs;
					meshes[m].glGroups[g].glFaces[f].glPairs = (GL_pair*)malloc(sizeof(GL_pair)*meshes[m].glGroups[g].glFaces[f].numglPairs);

					for(p = 0; p < meshes[m].jkFaces[j].numjkPairs; p++)
						{
						meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID = meshes[m].jkFaces[j].jkPairs[p].vertexID;
						meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID = meshes[m].jkFaces[j].jkPairs[p].txvertID;
						}

					f++;
					}
				}
			}

		// calculate glTxverts for groups with materials that had a file error
		for(g = 0; g < meshes[m].numglGroups; g++)
			if (materials[meshes[m].glGroups[g].matID].file_error)
				{
				maxS = maxT = 0.0f;

				for(f = 0; f < meshes[m].glGroups[g].numglFaces; f++)
					for(p = 0; p < meshes[m].glGroups[g].glFaces[f].numglPairs; p++)
						{
						if (meshes[m].jkTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s > maxS)
							maxS = meshes[m].jkTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s;
						if (meshes[m].jkTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t < maxT)
							maxT = meshes[m].jkTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t;
						}

				for(f = 0; f < meshes[m].glGroups[g].numglFaces; f++)
					for(p = 0; p < meshes[m].glGroups[g].glFaces[f].numglPairs; p++)
						{
						meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s = meshes[m].jkTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s / maxS;
						meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t = meshes[m].jkTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t / maxT;
						}
				}

		// reset the selection flag for all the glTxverts
		for(v = 0; v < meshes[m].numTxverts; v++)
			meshes[m].glTxverts[v].is_selected = false;

		// set the initial rendering mode for the mesh
		meshes[m].glRenderMode = SHOW;

		// print a warning if the mesh has too many faces
		if (meshes[m].numjkFaces > 512)
			{
			sprintf(warning, "Mesh %s contains %d surfaces.\n\nOnly 512 surfaces are allowed per mesh.\n\nThis 3do model will cause Jedi Knight to crash.", meshes[m].jkName, meshes[m].numjkFaces);
			MessageBox(ghwndMain, warning, "WARNING", MB_APPLMODAL);
			}
		}

	// use the min and max dimensions to center the model about the origin
	glViewOffset.x = (glMinX + glMaxX) / 2.0;
	glViewOffset.y = (glMinY + glMaxY) / 2.0;
	glViewOffset.z = (glMinZ + glMaxZ) / 2.0;

double temp;

	for(m = 0; m < numMeshes; m++)
		for(v = 0; v < meshes[m].numVerts; v++)
			{
			meshes[m].glVerts[v].x -= glViewOffset.x;
			meshes[m].glVerts[v].y -= glViewOffset.y;
			meshes[m].glVerts[v].z -= glViewOffset.z;

		temp = sqrt(meshes[m].glVerts[v].x*meshes[m].glVerts[v].x + meshes[m].glVerts[v].y*meshes[m].glVerts[v].y + meshes[m].glVerts[v].z*meshes[m].glVerts[v].z);

			if (temp > radius)
				radius = temp;
			}

	return true;
	}

void C3DOModel::Break_Txverts(int m)
	{
	CMatrix *table;
	int N;

	int cf, cp, cg;
	int tv, numTV;

	table = new CMatrix(meshes[m].numglGroups, meshes[m].numTxverts);

	// put a '1' at (row,col) if the group (row) uses the texturevertex (col).
	for(cf = 0; cf < meshes[m].numjkFaces; cf++)
		for(cp = 0; cp < meshes[m].jkFaces[cf].numjkPairs; cp++)
			table->Set(meshes[m].jkFaces[cf].groupID, meshes[m].jkFaces[cf].jkPairs[cp].txvertID, 1);

	// use the table to create extra texturevertices and make changes to each group as necessary.
	numTV = meshes[m].numTxverts;
	for(tv = 0; tv < numTV; tv++)
		{
		N = 0;
		for(cg = 0; cg < meshes[m].numglGroups; cg++)
			{
			N += table->Get(cg, tv);

			if (N > 1)
				{
				N = 1;

				meshes[m].jkTxverts = (JK_txvert*)realloc(meshes[m].jkTxverts, sizeof(JK_txvert)*(meshes[m].numTxverts +1));
				meshes[m].jkTxverts[meshes[m].numTxverts].s = meshes[m].jkTxverts[tv].s;
				meshes[m].jkTxverts[meshes[m].numTxverts].t = meshes[m].jkTxverts[tv].t;

				meshes[m].glTxverts = (GL_txvert*)realloc(meshes[m].glTxverts, sizeof(GL_txvert)*(meshes[m].numTxverts +1));
				meshes[m].glTxverts[meshes[m].numTxverts].s = meshes[m].glTxverts[tv].s;
				meshes[m].glTxverts[meshes[m].numTxverts].t = meshes[m].glTxverts[tv].t;

				for(cf = 0; cf < meshes[m].numjkFaces; cf++)
					if (meshes[m].jkFaces[cf].groupID == cg)
						for(cp = 0; cp < meshes[m].jkFaces[cf].numjkPairs; cp++)
							if (meshes[m].jkFaces[cf].jkPairs[cp].txvertID == tv)
								meshes[m].jkFaces[cf].jkPairs[cp].txvertID = meshes[m].numTxverts;

				meshes[m].numTxverts++;
				}
			}
		}

	delete table;
	}

bool C3DOModel::Save_File(char *szPathname)
	{
	int i, j, k;

	Prepare_Model();

	ofstream savefile(szPathname, ios::trunc, filebuf::sh_none);

	if (savefile.fail())
		return false;

	savefile << "# TEXTURED by " << gszAppName << endl
				<< endl
				<< "SECTION: HEADER" << endl
				<< endl
				<< "3DO " << jk3doVersion << endl
				<< endl
				<< "SECTION: MODELRESOURCE" << endl
				<< endl
				<< "MATERIALS " << numMaterials << endl;

	// save materials
	for(i = 0; i < numMaterials; i++)
		{
		savefile << i << ": " << materials[i].genericname << ".MAT" << endl;
		}

	savefile << endl
				<< "SECTION: GEOMETRYDEF" << endl
				<< endl
				<< "RADIUS " << jkModelRadius << endl
				<< endl
				<< "INSERT OFFSET " << jkInsertOffset.x << " " << jkInsertOffset.y << " " << jkInsertOffset.z << endl
				<< endl
				<< "GEOSETS " << "1" << endl // supports saving of only 1 geoset in this version.
				<< endl
				<< "GEOSET 0" << endl
				<< endl
				<< "MESHES " << numMeshes << endl;

	// save meshes
	for(i = 0; i < numMeshes; i++)
		{
		savefile << endl
					<< "MESH " << i << endl
					<< "NAME " << meshes[i].jkName << endl
					<< "RADIUS " << meshes[i].jkRadius << endl
					<< "GEOMETRYMODE " << meshes[i].jkGeometryMode << endl
					<< "LIGHTINGMODE " << meshes[i].jkLightingMode << endl
					<< "TEXTUREMODE " << meshes[i].jkTextureMode << endl
					<< endl
					<< "VERTICES " << meshes[i].numVerts << endl
					<< endl;

		// save vertex coordinates
		for(j = 0; j < meshes[i].numVerts; j++)
			{
			savefile << j << ": "
						<< meshes[i].jkVerts[j].x << " " << meshes[i].jkVerts[j].y << " "
						<< meshes[i].jkVerts[j].z << " " << meshes[i].jkVerts[j].i << endl;
			}

//////////////////////////// save texture vertex coordinates
		savefile << endl
					<< "TEXTURE VERTICES " << meshes[i].numOutputTxverts << endl
					<< endl;

		for(j = 0; j < meshes[i].numOutputTxverts; j++)
			{
			savefile << j << ": "
						<< meshes[i].outputTxverts[j].s << " " << meshes[i].outputTxverts[j].t << endl;
			}

		savefile << endl
					<< "VERTEX NORMALS" << endl
					<< endl;

		// save vertex normal coordinates
		for(j = 0; j < meshes[i].numVerts; j++)
			{
			savefile << j << ": " << meshes[i].jkVertnormals[j].x << " "
						<< meshes[i].jkVertnormals[j].y << " " << meshes[i].jkVertnormals[j].z << endl;
			}

///////////////////////////////////// save faces information
		savefile << endl
					<< "FACES " << meshes[i].numjkFaces << endl
					<< endl;

		for(j = 0; j < meshes[i].numjkFaces; j++)
			{
			savefile << j << ": " << meshes[i].outputFaces[j].matID << " "
						<< meshes[i].outputFaces[j].jkType << " " << meshes[i].outputFaces[j].jkGeo << " "
						<< meshes[i].outputFaces[j].jkLight << " " << meshes[i].outputFaces[j].jkTex << " "
						<< meshes[i].outputFaces[j].jkExtralight << " "
						<< meshes[i].outputFaces[j].numjkPairs;

			for(k = 0; k < meshes[i].outputFaces[j].numjkPairs; k++)
				{
				savefile << " "  << meshes[i].outputFaces[j].jkPairs[k].vertexID
							<< ", " << meshes[i].outputFaces[j].jkPairs[k].txvertID;
				}

			savefile << endl;
			}

		savefile << endl
					<< "FACE NORMALS" << endl
					<< endl;

		// save face normal coordinates
		for(j = 0; j < meshes[i].numjkFaces; j++)
			{
			savefile << j << ": "
						<< meshes[i].jkFacenormals[meshes[i].outputFaces[j].jkFaceID].x << " "
						<< meshes[i].jkFacenormals[meshes[i].outputFaces[j].jkFaceID].y << " "
						<< meshes[i].jkFacenormals[meshes[i].outputFaces[j].jkFaceID].z << endl;
			}
		} // end mesh loop

	savefile << endl
				<< "SECTION: HIERARCHYDEF" << endl
				<< endl
				<< "HIERARCHY NODES " << numHNodes << endl
				<< "# num: flags: type: mesh: parent: child: sibling: numChildren: x: y: z: pitch: yaw: roll: pivotx: pivoty: pivotz: hnodename:" << endl;

	// save node hierarchy
	for(i = 0; i < numHNodes; i++)
		{
		savefile << i << ": " << hNodes[i].flags << " " << hNodes[i].type << " " << hNodes[i].mesh << " "
					<< hNodes[i].parent << " " << hNodes[i].child << " " << hNodes[i].sibling << " " << hNodes[i].numChildren << " "
					<< hNodes[i].x << " " << hNodes[i].y << " " << hNodes[i].z << " "
					<< hNodes[i].pitch << " " << hNodes[i].yaw << " " << hNodes[i].roll << " "
					<< hNodes[i].pivotX << " " << hNodes[i].pivotY << " " << hNodes[i].pivotZ << " "
					<< hNodes[i].hNodeName << endl;
		}

	savefile.close();

	return true;
	}

void C3DOModel::Prepare_Model()
	{
	int m, v, g, f, p;
	int curOutputTxvert, curOutputFace;

	for(m = 0; m < numMeshes; m++)
		{
		meshes[m].numOutputTxverts = 0;
		free(meshes[m].outputTxverts);

		// reset txvert flags
		for(v = 0; v < meshes[m].numTxverts; v++)
			meshes[m].glTxverts[v].flag = false;

		// determine the number of txverts to be exported
		for(g = 0; g < meshes[m].numglGroups; g++)
			{
			if (true == meshes[m].glGroups[g].is_original)
				{
				for(f = 0; f < meshes[m].glGroups[g].numglFaces; f++)
					for(p = 0; p < meshes[m].glGroups[g].glFaces[f].numglPairs; p++)
						if (false == meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].flag)
							{
							meshes[m].numOutputTxverts ++;
							meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].flag = true;
							}
				}
			else
				meshes[m].numOutputTxverts += meshes[m].glGroups[g].numglTxverts;
			}

		// allocate memory for output txverts
		meshes[m].outputTxverts = (JK_txvert*)malloc(sizeof(JK_txvert) * meshes[m].numOutputTxverts);

		// reset txvert flags
		for(v = 0; v < meshes[m].numTxverts; v++)
			meshes[m].glTxverts[v].redirect = -1;

		// generate the output txverts
		curOutputTxvert = -1;
		for(g = 0; g < meshes[m].numglGroups; g++)
			{
			if (true == meshes[m].glGroups[g].is_original)
				{
				for(f = 0; f < meshes[m].glGroups[g].numglFaces; f++)
					for(p = 0; p < meshes[m].glGroups[g].glFaces[f].numglPairs; p++)
						if (-1 == meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].redirect)
							{
							curOutputTxvert ++;

							if (! materials[meshes[m].glGroups[g].matID].file_error)
								{
								meshes[m].outputTxverts[curOutputTxvert].s =   meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s * materials[meshes[m].glGroups[g].matID].width;
								meshes[m].outputTxverts[curOutputTxvert].t = - meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t * materials[meshes[m].glGroups[g].matID].height;
								}
							else
								{
								meshes[m].outputTxverts[curOutputTxvert].s =   meshes[m].jkTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s;
								meshes[m].outputTxverts[curOutputTxvert].t =   meshes[m].jkTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t;
								}

							meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].redirect = curOutputTxvert;
							}
				}
			else
				{
				for(v = 0; v < meshes[m].glGroups[g].numglTxverts; v++)
					{
					curOutputTxvert ++;

					if (! materials[meshes[m].glGroups[g].matID].file_error)
						{
						meshes[m].outputTxverts[curOutputTxvert].s =   meshes[m].glGroups[g].glTxverts[v].s * materials[meshes[m].glGroups[g].matID].width;
						meshes[m].outputTxverts[curOutputTxvert].t = - meshes[m].glGroups[g].glTxverts[v].t * materials[meshes[m].glGroups[g].matID].height;
						}
					else
						{
						meshes[m].outputTxverts[curOutputTxvert].s =   meshes[m].glGroups[g].glTxverts[v].s;
						meshes[m].outputTxverts[curOutputTxvert].t = - meshes[m].glGroups[g].glTxverts[v].t;
						}

					meshes[m].glGroups[g].glTxverts[v].redirect = curOutputTxvert;
					}
				}
			}

		// allocate memory for the output faces
		meshes[m].outputFaces = (JK_face *)malloc(sizeof(JK_face) * meshes[m].numjkFaces);

		// generate the output faces
		curOutputFace = -1;
		for(g = 0; g < meshes[m].numglGroups; g++)
			for(f = 0; f < meshes[m].glGroups[g].numglFaces; f++)
				{
				curOutputFace ++;
				meshes[m].outputFaces[curOutputFace].matID = meshes[m].glGroups[g].matID;

				strcpy(meshes[m].outputFaces[curOutputFace].jkType, meshes[m].jkFaces[meshes[m].glGroups[g].glFaces[f].jkFaceID].jkType);
				meshes[m].outputFaces[curOutputFace].jkGeo        = meshes[m].jkFaces[meshes[m].glGroups[g].glFaces[f].jkFaceID].jkGeo;
				meshes[m].outputFaces[curOutputFace].jkLight      = meshes[m].jkFaces[meshes[m].glGroups[g].glFaces[f].jkFaceID].jkLight;
				meshes[m].outputFaces[curOutputFace].jkTex        = meshes[m].jkFaces[meshes[m].glGroups[g].glFaces[f].jkFaceID].jkTex;
				meshes[m].outputFaces[curOutputFace].jkExtralight = meshes[m].jkFaces[meshes[m].glGroups[g].glFaces[f].jkFaceID].jkExtralight;

				meshes[m].outputFaces[curOutputFace].jkFaceID     = meshes[m].glGroups[g].glFaces[f].jkFaceID;

				meshes[m].outputFaces[curOutputFace].numjkPairs = meshes[m].glGroups[g].glFaces[f].numglPairs;
				meshes[m].outputFaces[curOutputFace].jkPairs = (JK_pair *)malloc(sizeof(JK_pair) * meshes[m].outputFaces[curOutputFace].numjkPairs);

				for(p = 0; p < meshes[m].glGroups[g].glFaces[f].numglPairs; p++)
					{
					meshes[m].outputFaces[curOutputFace].jkPairs[p].vertexID = meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID;
					if (true == meshes[m].glGroups[g].is_original)
						meshes[m].outputFaces[curOutputFace].jkPairs[p].txvertID = meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].redirect;
					else
						meshes[m].outputFaces[curOutputFace].jkPairs[p].txvertID = meshes[m].glGroups[g].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].redirect;
					}
				}
		}
	}

void C3DOModel::Display()
	{
	int m, g, f, p;

	glPushMatrix();

		glRotated(gCamera->getVAngle(), -1.0, 0.0, 0.0);
		glRotated(gCamera->getHAngle(), 0.0, -1.0, 0.0);

		glScaled(gCamera->getScaleFactor(), gCamera->getScaleFactor(), gCamera->getScaleFactor());

		for(m = 0; m < numMeshes; m++)
			{
			for(g = 0; g < meshes[m].numglGroups; g++)
				{
				// if the group is selected
				if (meshes[m].glGroups[g].is_selected)
					{
					glColor3f(0.8f, 1.0f, 0.8f);
					glEnable(GL_BLEND);
					glEnable(GL_TEXTURE_2D);
					glEnable(GL_CULL_FACE);
					glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
					glBindTexture(GL_TEXTURE_2D, materials[meshes[m].glGroups[g].matID].glTexture[0]);
					}
				// if the group is not selected and the mesh is shown
				else if (meshes[m].glRenderMode == SHOW)
					{
					glColor3f(1.0, 1.0, 1.0);
					glEnable(GL_BLEND);
					glEnable(GL_TEXTURE_2D);
					glEnable(GL_CULL_FACE);
					glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
					glBindTexture(GL_TEXTURE_2D, materials[meshes[m].glGroups[g].matID].glTexture[0]);
					}
				// if the group is not selected and the mesh is wireframe
				else if (meshes[m].glRenderMode == WIRE)
					{
					glColor3f(1.0, 1.0, 1.0);
					glEnable(GL_BLEND);
					glDisable(GL_TEXTURE_2D);
					glDisable(GL_CULL_FACE);
					glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
					}

				// if the group is selected and/or the mesh is not hidden, then display it
				if ((meshes[m].glGroups[g].is_selected) || (meshes[m].glRenderMode != HIDE))
					{
					// if the group has not been remapped, then use the original txverts contained by the mesh
					if (meshes[m].glGroups[g].is_original)
						{
						for(f = 0; f < meshes[m].glGroups[g].numglFaces; f++)
							{
							glBegin(GL_POLYGON);
								for(p = 0; p < meshes[m].glGroups[g].glFaces[f].numglPairs; p++)
									{
									glTexCoord2d(meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s,
													 meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t);

									glVertex3d(meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].x,
												  meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].z,
												  -meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].y);
									}
							glEnd();
							}
						}

					/*******************/

					// if the group has been remapped, then use the new txverts contained within the group
					else
						{
						for(f = 0; f < meshes[m].glGroups[g].numglFaces; f++)
							{
							glBegin(GL_POLYGON);
								for(p = 0; p < meshes[m].glGroups[g].glFaces[f].numglPairs; p++)
									{
									glTexCoord2d(meshes[m].glGroups[g].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s,
													 meshes[m].glGroups[g].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t);

									glVertex3d(meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].x,
												  meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].z,
												  -meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].y);
									}
							glEnd();
							}
						}
					}

				// finally, if the group is selected then trace a green wireframe around it
				if (meshes[m].glGroups[g].is_selected)
					{
					glColor3f(0.8f, 1.0f, 0.8f);
					glDisable(GL_BLEND);
					glDisable(GL_TEXTURE_2D);
					glDisable(GL_CULL_FACE);
					glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

					// if the group has not been remapped, then use the original txverts contained by the mesh
					if (meshes[m].glGroups[g].is_original)
						{
						for(f = 0; f < meshes[m].glGroups[g].numglFaces; f++)
							{
							glBegin(GL_POLYGON);
								for(p = 0; p < meshes[m].glGroups[g].glFaces[f].numglPairs; p++)
									{
									glTexCoord2d(meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s,
													 meshes[m].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t);

									glVertex3d(meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].x,
												  meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].z,
												  -meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].y);
									}
							glEnd();
							}
						}

					/*******************/

					// if the group has been remapped, then use the new txverts contained within the group
					else
						{
						for(f = 0; f < meshes[m].glGroups[g].numglFaces; f++)
							{
							glBegin(GL_POLYGON);
								for(p = 0; p < meshes[m].glGroups[g].glFaces[f].numglPairs; p++)
									{
									glTexCoord2d(meshes[m].glGroups[g].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].s,
													 meshes[m].glGroups[g].glTxverts[meshes[m].glGroups[g].glFaces[f].glPairs[p].txvertID].t);

									glVertex3d(meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].x,
												  meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].z,
												  -meshes[m].glVerts[meshes[m].glGroups[g].glFaces[f].glPairs[p].vertexID].y);
									}
							glEnd();
							}
						}
					}
				}
			}

	glPopMatrix();
	}

void C3DOModel::Deselect_AllMeshes()
	{
	for(int m = 0; m < numMeshes; m++)
		meshes[m].Deselect_AllGroups();
	}

void C3DOModel::Change_Tex(int sel, int sel2, char *szTitlename, char *szPathname)
	{
	int oldMatID, newMatID;
	int t, m, g, numUsing;
	bool referencedToExisting;
	char genericName[_MAX_FNAME];
	char ext;

	strcpy(genericName, szTitlename);
	genericName[strlen(genericName)-4] = '\0';
	ext = szPathname[strlen(szPathname)-1];

	oldMatID = meshes[sel].glGroups[sel2].matID;
	newMatID = -1;
	referencedToExisting = false;

	// find out if the texture is already in the materials list.
	// if so, reference this group to that material, and perform a consistency check at the end.
	for(t = 0; t < numMaterials; t++)
		if (!(strcmpi(szPathname, materials[t].pathname)))
			{
			newMatID = t;
			referencedToExisting = true;
			t = numMaterials;
			}

	// if the material was not already in the list,
	// determine if the material being replaced will no longer be used, and if so replace that one.
	if (!referencedToExisting)
		{
		numUsing = 0;
		for(m = 0; m < numMeshes; m++)
			for(g = 0; g < meshes[m].numglGroups; g++)
				if (meshes[m].glGroups[g].matID == oldMatID)
					numUsing ++;

		if (numUsing == 1)
			newMatID = oldMatID;
		}

	// if both of those possibilities failed, then we must create a new material.
	if (newMatID == -1)
		{
		newMatID = numMaterials;

		if(!(materials = (CMaterial*)realloc(materials, sizeof(CMaterial) * (numMaterials +1))))
			MessageBox(ghwndMain, "Could not allocate memory for: materials", "Error", MB_APPLMODAL);

		numMaterials++;
		}

	// now, create the GL texture (even if the same MAT/BMP file was selected).
	materials[newMatID].Create_glTexture(szPathname);

	// update the info held by the group.
	meshes[sel].glGroups[sel2].matID = newMatID;
	strcpy(meshes[sel].glGroups[sel2].glGroupName, genericName);

	// now, if the group was referenced to an existing material,
	// we have to check if the old material is now left unused.
	if (referencedToExisting)
		{
		numUsing = 0;
		for(m = 0; m < numMeshes; m++)
			for(g = 0; g < meshes[m].numglGroups; g++)
				if (meshes[m].glGroups[g].matID == oldMatID)
					numUsing ++;

		// if it was indeed left unused, then we have to do a work-around
		if (numUsing == 0)
			{
			// if the material was the last one in the list, simply reduce numMaterials by 1.
			if (oldMatID == numMaterials-1)
				numMaterials --;
			// otherwise, do the work-around.
			else
				{
				// copy the last material into the one that is now unused.
				strcpy(materials[oldMatID].pathname, materials[numMaterials-1].pathname);

				materials[oldMatID].Create_glTexture(materials[oldMatID].pathname);

				// update the info held by any group that references the last material.
				for(m = 0; m < numMeshes; m++)
					for(g = 0; g < meshes[m].numglGroups; g++)
						if (meshes[m].glGroups[g].matID == numMaterials-1)
							meshes[m].glGroups[g].matID = oldMatID;

				// now delete the last material
				numMaterials --;
				}
			}
		}
	}

//////////////////////////////////////////////////////////////////////////////////////////// CUVMap

class CUVMap
	{
	public:

	int meshID;
	int groupID;
	int texID;

	bool selecting;
	double anchorX, anchorY, cursorX, cursorY;
	double selTop, selBottom, selLeft, selRight;

	double minS, ctrS, maxS;
	double minT, ctrT, maxT;
	double viewRad;

	CUVMap(int, int);

	void Compute_Dimensions();
	void SetupOrthoView();
	void Display();

	void Remap(int);

	bool TestForSelection();
	void MoveSelection(double, double);
	void ScaleX(double);
	void ScaleY(double);

	void Select_Point(double, double);
	void Select_RectStart(double, double);
	void Select_RectSize(double, double);
	void Select_RectEnd();

	void Select_All();
	void Select_None();
	void Select_Invert();
	};

CUVMap::CUVMap(int m, int g)
	{
	meshID = m;
	groupID = g;
	texID = gmodel->meshes[m].glGroups[g].matID;

	selecting = false;

	Compute_Dimensions();
	}

void CUVMap::Compute_Dimensions()
	{
	int f, p, v;
	double curS, curT;

	minS = minT = 1.7e300;
	maxS = maxT = -1.7e300;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
		for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
			{
			v = gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID;

			curS = gmodel->meshes[meshID].glTxverts[v].s;
			curT = gmodel->meshes[meshID].glTxverts[v].t;

			if (curS < minS) minS = curS;
			if (curS > maxS) maxS = curS;
			if (curT < minT) minT = curT;
			if (curT > maxT) maxT = curT;
			}

	////////////////

	else
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			{
			curS = gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s;
			curT = gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t;

			if (curS < minS) minS = curS;
			if (curS > maxS) maxS = curS;
			if (curT < minT) minT = curT;
			if (curT > maxT) maxT = curT;
			}

	ctrS = (maxS + minS) / 2.0;
	ctrT = (maxT + minT) / 2.0;

	viewRad = max((maxS - ctrS), (maxT - ctrT));
	}

void CUVMap::SetupOrthoView()
	{
	glViewport(0, 0, grectGLV.right, grectGLV.bottom);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glOrtho(-viewRad, viewRad, -viewRad, viewRad, 0.0, 10.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	}

void CUVMap::Display()
	{
	int cf, cp, ct;
	int row, col;

	gluLookAt(ctrS, ctrT, 5.0, ctrS, ctrT, 0.0, 0.0, 1.0, 0.0);

	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	glBindTexture(GL_TEXTURE_2D, gmodel->materials[texID].glTexture[0]);
	glColor3f(1.0, 1.0, 1.0);

	for(row = (int)minT-1; row < (int)maxT+1; row++)
	for(col = (int)minS-1; col < (int)maxS+1; col++)
		{
		glBegin(GL_POLYGON);
			// lower left corner
			glTexCoord2d(0.0, 0.0);
			glVertex3d((double)col, (double)row, -3.0);

			// lower right corner
			glTexCoord2d(1.0, 0.0);
			glVertex3d((double)col+1.0, (double)row, -3.0);

			// upper right corner
			glTexCoord2d(1.0, 1.0);
			glVertex3d((double)col+1.0, (double)row+1.0, -3.0);

			// upper left corner
			glTexCoord2d(0.0, 1.0);
			glVertex3d((double)col, (double)row+1.0, -3.0);
		glEnd();
		}

	////////////////////////////////////////

	// display the connecting edges and GL texture vertices
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	glLineWidth(1.0);
	glPointSize(5.0);

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		{
		for(cf = 0; cf < gmodel->meshes[meshID].glGroups[groupID].numglFaces; cf++)
			{
			// display the connecting edges
			glColor3f(1.0f, 1.0f, 1.0f);

			glBegin(GL_POLYGON);
				for(cp = 0; cp < gmodel->meshes[meshID].glGroups[groupID].glFaces[cf].numglPairs; cp++)
					{
					ct = gmodel->meshes[meshID].glGroups[groupID].glFaces[cf].glPairs[cp].txvertID;

					glVertex3d(gmodel->meshes[meshID].glTxverts[ct].s, gmodel->meshes[meshID].glTxverts[ct].t, -2.0);
					}
			glEnd();

			// display the selected and non-selected GL texture vertices
			for(cp = 0; cp < gmodel->meshes[meshID].glGroups[groupID].glFaces[cf].numglPairs; cp++)
				{
				ct = gmodel->meshes[meshID].glGroups[groupID].glFaces[cf].glPairs[cp].txvertID;

				if (false == gmodel->meshes[meshID].glTxverts[ct].is_selected)
					glColor3f(1.0f, 0.0f, 0.0f);
				else
					glColor3f(0.0f, 1.0f, 0.0f);

				glBegin(GL_POINTS);
					glVertex3d(gmodel->meshes[meshID].glTxverts[ct].s, gmodel->meshes[meshID].glTxverts[ct].t, -1.0);
				glEnd();
				}
			}
		}

	////////////////

	else
		{
		for(cf = 0; cf < gmodel->meshes[meshID].glGroups[groupID].numglFaces; cf++)
			{
			// display the connecting edges
			glColor3f(1.0f, 1.0f, 1.0f);

			glBegin(GL_POLYGON);
				for(cp = 0; cp < gmodel->meshes[meshID].glGroups[groupID].glFaces[cf].numglPairs; cp++)
					{
					ct = gmodel->meshes[meshID].glGroups[groupID].glFaces[cf].glPairs[cp].txvertID;

					glVertex3d(gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].s, gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].t, -2.0);
					}
			glEnd();

			// display the selected and non-selected GL texture vertices
			for(cp = 0; cp < gmodel->meshes[meshID].glGroups[groupID].glFaces[cf].numglPairs; cp++)
				{
				ct = gmodel->meshes[meshID].glGroups[groupID].glFaces[cf].glPairs[cp].txvertID;

				if (false == gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].is_selected)
					glColor3f(1.0f, 0.0f, 0.0f);
				else
					glColor3f(0.0f, 1.0f, 0.0f);

				glBegin(GL_POINTS);
					glVertex3d(gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].s, gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].t, -1.0);
				glEnd();
				}
			}
		}

	if (selecting)
		{
		// display the selection box
		glColor3f(0.0f, 1.0f, 0.0f);

		glBegin(GL_LINE_LOOP);
			glVertex3d(selLeft,  selBottom, 0.0);
			glVertex3d(selRight, selBottom, 0.0);
			glVertex3d(selRight, selTop,    0.0);
			glVertex3d(selLeft,  selTop,    0.0);
		glEnd();
		}
	}

void CUVMap::Remap(int mode)
	{
	int v, f, p, ct;

	double curS, curT;
	double rminS, rminT;
	double rmaxS, rmaxT;

	rminS = rminT =  1.7e300;
	rmaxS = rmaxT = -1.7e300;

	// if the group hasn't yet been remapped...
	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		{
		// select none of the mesh's txverts just to be safe
		Select_None();

		// reset algorithm flags
		for(v = 0; v < gmodel->meshes[meshID].numVerts; v++)
			gmodel->meshes[meshID].glVerts[v].flag = false;

		// compute number of glTxverts to be generated (the number of vertices that are referenced by the group)
		gmodel->meshes[meshID].glGroups[groupID].numglTxverts = 0;
		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
			for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
				if (false == gmodel->meshes[meshID].glVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].flag)
					{
					gmodel->meshes[meshID].glGroups[groupID].numglTxverts ++;
					gmodel->meshes[meshID].glVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].flag = true;
					}

		// erase the old glTxverts and allocate the space for the new ones
		if (gmodel->meshes[meshID].glGroups[groupID].glTxverts != NULL)
			free(gmodel->meshes[meshID].glGroups[groupID].glTxverts);
		gmodel->meshes[meshID].glGroups[groupID].glTxverts = (GL_txvert*)malloc(sizeof(GL_txvert)*gmodel->meshes[meshID].glGroups[groupID].numglTxverts);
		if (gmodel->meshes[meshID].glGroups[groupID].glTxverts == NULL)
			MessageBox(ghwndMain, "Could not allocate memory for: glTxverts", "Error", MB_APPLMODAL);

		// the group is no longer original as specified in the 3do file
		gmodel->meshes[meshID].glGroups[groupID].is_original = false;

		// select none of the new txverts in the group
		Select_None();
		}

	////////////////

	// reset algorithm flags
	for(v = 0; v < gmodel->meshes[meshID].numVerts; v++)
		gmodel->meshes[meshID].jkVerts[v].temp = -1;

	// compute the group's new txverts based on the direction of the orthographic mapping
	ct = 0;
	for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
		for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
			{
			if (-1 == gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].temp)
				{
				if (mode == ID_UVORTHOTOP)
					{
					curS = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].s =   gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].x;
					curT = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].t =   gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].y;
					}
				else if (mode == ID_UVORTHOBOTTOM)
					{
					curS = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].s = - gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].x;
					curT = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].t =   gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].y;
					}
				else if (mode == ID_UVORTHOFRONT)
					{
					curS = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].s = - gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].x;
					curT = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].t =   gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].z;
					}
				else if (mode == ID_UVORTHOBACK)
					{
					curS = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].s =   gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].x;
					curT = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].t =   gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].z;
					}
				else if (mode == ID_UVORTHOLEFT)
					{
					curS = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].s =   gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].y;
					curT = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].t =   gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].z;
					}
				else if (mode == ID_UVORTHORIGHT)
					{
					curS = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].s = - gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].y;
					curT = gmodel->meshes[meshID].glGroups[groupID].glTxverts[ct].t =   gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].z;
					}

				if (curS < rminS) rminS = curS;
				if (curT < rminT) rminT = curT;

				gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID = ct;

				gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].temp = ct;
				ct++;
				}
			else
				{
				gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID = gmodel->meshes[meshID].jkVerts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].vertexID].temp;
				}
			}

	// translate all txverts to upper-right of origin
	for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
		{
		curS = gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s -= rminS;
		curT = gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t -= rminT;

		if (curS > rmaxS) rmaxS = curS;
		if (curT > rmaxT) rmaxT = curT;
		}

	// scale all txverts to fill the texture
	for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
		{
		gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s /= rmaxS;
		gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t /= rmaxT;
		}

	Compute_Dimensions();
	if (gbUVMapMode_Active)
		{
		SetupOrthoView();
		Display();
		}
	}

bool CUVMap::TestForSelection()
	{
	int f, p, v;
	bool selection = false;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		{
		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
		for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
			{
			v = gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID;

			if (gmodel->meshes[meshID].glTxverts[v].is_selected)
				{
				selection = true;
				f = gmodel->meshes[meshID].glGroups[groupID].numglFaces;
				p = gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs;
				}
			}
		}

	////////////////

	else
		{
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			if (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected)
				{
				selection = true;
				v = gmodel->meshes[meshID].glGroups[groupID].numglTxverts;
				}
		}

	if (!selection)
		Msg(ghwndControls, (char*)gszAppName, "You must make a selection first.\nDrag a rectangle or click on individual vertices to make a selection.");

	return selection;
	}

void CUVMap::MoveSelection(double cx, double cy)
	{
	int f, p, v;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		{
		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			gmodel->meshes[meshID].glTxverts[v].flag = false;

		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
		for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
			{
			v = gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID;

			if (false == gmodel->meshes[meshID].glTxverts[v].flag)
				{
				if (true == gmodel->meshes[meshID].glTxverts[v].is_selected)
					{
					gmodel->meshes[meshID].glTxverts[v].s += cx;
					gmodel->meshes[meshID].glTxverts[v].t += cy;
					}

				gmodel->meshes[meshID].glTxverts[v].flag = true;
				}
			}
		}

	////////////////

	else
		{
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			if (true == gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected)
				{
				gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s += cx;
				gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t += cy;
				}
		}

	Compute_Dimensions();
	SetupOrthoView();
	Display();
	}

void CUVMap::ScaleX(double amount)
	{
	int f, p, v;
	double curS, SminS, SmaxS, SavgS;

	SminS =  1.7e300;
	SmaxS = -1.7e300;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		{
		// reset algorithm flags
		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			gmodel->meshes[meshID].glTxverts[v].flag = false;

		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
			for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
				gmodel->meshes[meshID].glTxverts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID].flag = true;

		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			if ((gmodel->meshes[meshID].glTxverts[v].is_selected) && (gmodel->meshes[meshID].glTxverts[v].flag))
				{
				curS = gmodel->meshes[meshID].glTxverts[v].s;

				if (curS < SminS) SminS = curS;
				if (curS > SmaxS) SmaxS = curS;
				}

		SavgS = (SmaxS + SminS) / 2.0;

		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			if ((gmodel->meshes[meshID].glTxverts[v].is_selected) && (gmodel->meshes[meshID].glTxverts[v].flag))
				{
				gmodel->meshes[meshID].glTxverts[v].s -= SavgS;
				gmodel->meshes[meshID].glTxverts[v].s *= amount;
				gmodel->meshes[meshID].glTxverts[v].s += SavgS;

				gmodel->meshes[meshID].glTxverts[v].flag = false;
				}
		}

	////////////////

	else
		{
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			if (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected)
				{
				curS = gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s;

				if (curS < SminS) SminS = curS;
				if (curS > SmaxS) SmaxS = curS;
				}

		SavgS = (SmaxS + SminS) / 2.0;

		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			if (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected)
				{
				gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s -= SavgS;
				gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s *= amount;
				gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s += SavgS;
				}
		}

	Compute_Dimensions();
	SetupOrthoView();
	Display();
	}

void CUVMap::ScaleY(double amount)
	{
	int f, p, v;
	double curT, SminT, SmaxT, SavgT;

	SminT =  1.7e300;
	SmaxT = -1.7e300;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		{
		// reset algorithm flags
		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			gmodel->meshes[meshID].glTxverts[v].flag = false;

		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
			for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
				gmodel->meshes[meshID].glTxverts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID].flag = true;

		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			if ((gmodel->meshes[meshID].glTxverts[v].is_selected) && (gmodel->meshes[meshID].glTxverts[v].flag))
				{
				curT = gmodel->meshes[meshID].glTxverts[v].t;

				if (curT < SminT) SminT = curT;
				if (curT > SmaxT) SmaxT = curT;
				}

		SavgT = (SmaxT + SminT) / 2.0;

		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			if ((gmodel->meshes[meshID].glTxverts[v].is_selected) && (gmodel->meshes[meshID].glTxverts[v].flag))
				{
				gmodel->meshes[meshID].glTxverts[v].t -= SavgT;
				gmodel->meshes[meshID].glTxverts[v].t *= amount;
				gmodel->meshes[meshID].glTxverts[v].t += SavgT;

				gmodel->meshes[meshID].glTxverts[v].flag = false;
				}
		}

	////////////////

	else
		{
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			if (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected)
				{
				curT = gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t;

				if (curT < SminT) SminT = curT;
				if (curT > SmaxT) SmaxT = curT;
				}

		SavgT = (SmaxT + SminT) / 2.0;

		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			if (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected)
				{
				gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t -= SavgT;
				gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t *= amount;
				gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t += SavgT;
				}
		}

	Compute_Dimensions();
	SetupOrthoView();
	Display();
	}

void CUVMap::Select_Point(double x, double y)
	{
	int f, p, v;
	double minX, maxX, minY, maxY;

	minX = x -0.003;
	maxX = x +0.003;
	minY = y -0.003;
	maxY = y +0.003;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		{
		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			gmodel->meshes[meshID].glTxverts[v].flag = false;

		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
		for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
				{
				v = gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID;

				if (false == gmodel->meshes[meshID].glTxverts[v].flag)
					if ((gmodel->meshes[meshID].glTxverts[v].s > minX) && (gmodel->meshes[meshID].glTxverts[v].s < maxX) && (gmodel->meshes[meshID].glTxverts[v].t > minY) && (gmodel->meshes[meshID].glTxverts[v].t < maxY))
						{
						gmodel->meshes[meshID].glTxverts[v].is_selected = !gmodel->meshes[meshID].glTxverts[v].is_selected;
						gmodel->meshes[meshID].glTxverts[v].flag = true;
						}
				}
		}

	////////////////

	else
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			if ((gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s > minX) && (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s < maxX) && (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t > minY) && (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t < maxY))
				gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected = !gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected;
	}

void CUVMap::Select_RectStart(double x, double y)
	{
	selecting = true;

	anchorX = cursorX = selLeft = selRight = x;
	anchorY = cursorY = selTop = selBottom = y;
	}

void CUVMap::Select_RectSize(double cx, double cy)
	{
	cursorX += cx;
	cursorY += cy;

	if (cursorY > anchorY)        { selTop = cursorY; selBottom = anchorY; }
	else if (cursorY < anchorY)   { selTop = anchorY; selBottom = cursorY; }
	else                          { selTop = selBottom = anchorY;          }

	if (cursorX > anchorX)        { selRight = cursorX; selLeft = anchorX; }
	else if (cursorX < anchorX)   { selRight = anchorX; selLeft = cursorX; }
	else                          { selRight = selLeft = anchorX;          }
	}

void CUVMap::Select_RectEnd()
	{
	int f, p, v;

	selecting = false;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		{
		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			gmodel->meshes[meshID].glTxverts[v].flag = false;

		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
		for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
				{
				v = gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID;

				if (false == gmodel->meshes[meshID].glTxverts[v].flag)
					if ((gmodel->meshes[meshID].glTxverts[v].s > selLeft) && (gmodel->meshes[meshID].glTxverts[v].s < selRight) && (gmodel->meshes[meshID].glTxverts[v].t > selBottom) && (gmodel->meshes[meshID].glTxverts[v].t < selTop))
						{
						gmodel->meshes[meshID].glTxverts[v].is_selected = !gmodel->meshes[meshID].glTxverts[v].is_selected;
						gmodel->meshes[meshID].glTxverts[v].flag = true;
						}
				}
		}

	////////////////

	else
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			if ((gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s > selLeft) && (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].s < selRight) && (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t > selBottom) && (gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].t < selTop))
					gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected = !gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected;
	}

void CUVMap::Select_All()
	{
	int f, p, v;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
		for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
			gmodel->meshes[meshID].glTxverts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID].is_selected = true;

	////////////////

	else
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected = true;
	}

void CUVMap::Select_None()
	{
	int f, p, v;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
		for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
			gmodel->meshes[meshID].glTxverts[gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID].is_selected = false;

	////////////////

	else
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected = false;
	}

void CUVMap::Select_Invert()
	{
	int f, p, v;

	if (true == gmodel->meshes[meshID].glGroups[groupID].is_original)
		{
		for(v = 0; v < gmodel->meshes[meshID].numTxverts; v++)
			gmodel->meshes[meshID].glTxverts[v].flag = false;

		for(f = 0; f < gmodel->meshes[meshID].glGroups[groupID].numglFaces; f++)
		for(p = 0; p < gmodel->meshes[meshID].glGroups[groupID].glFaces[f].numglPairs; p++)
			{
			v = gmodel->meshes[meshID].glGroups[groupID].glFaces[f].glPairs[p].txvertID;

			if (false == gmodel->meshes[meshID].glTxverts[v].flag)
				{
				gmodel->meshes[meshID].glTxverts[v].is_selected = !gmodel->meshes[meshID].glTxverts[v].is_selected;
				gmodel->meshes[meshID].glTxverts[v].flag = true;
				}
			}
		}

	////////////////

	else
		for(v = 0; v < gmodel->meshes[meshID].glGroups[groupID].numglTxverts; v++)
			gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected = !gmodel->meshes[meshID].glGroups[groupID].glTxverts[v].is_selected;
	}

/////////////////////////////////////////////////////////////////////////////// GL_RenderScene

void GL_RenderScene()
	{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	if (gbModelMode_Active)
		{
		gCamera->doLookAt();
		gmodel->Display();
		}
	else if (gbUVMapMode_Active)
		{
		guvmap->Display();
		}

	glFlush();
	SwapBuffers(ghdcGLV);         // bring backbuffer to foreground
	}

///////////////////////////////////////////////////////////////////////////// GL_InitializeSettings

void GL_InitializeSettings()
	{
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);              // clear to black

	glShadeModel(GL_SMOOTH);                           // use smooth shading
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_DEPTH_TEST);                           // hidden surface removal

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // Really Nice Perspective Calculations

	/*glEnable(GL_LIGHTING);                           // setup cheap-o lighting
	glEnable(GL_LIGHT0);
	glEnable(GL_COLOR_MATERIAL);*/

	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POLYGON_SMOOTH);

	glHint(GL_POINT_SMOOTH, GL_NICEST);
	glHint(GL_LINE_SMOOTH, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH, GL_NICEST);

	glEnable(GL_CULL_FACE);                            // do not calculate inside of polys
	glFrontFace(GL_CCW);                               // counterclockwise polys are facing out
	}

////////////////////////////////////////////////////////////////////////// GL_InitializePixelFormat

// desc: sets the pixel format for the device context
void GL_InitializePixelFormat(HDC hDC)
	{
	int nPixelFormat;                   // our pixel format index

	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),   // size of structure
		1,                               // default version
		PFD_DRAW_TO_WINDOW |             // window drawing support
		PFD_SUPPORT_OPENGL |             // OpenGL support
		PFD_DOUBLEBUFFER,                // double buffering support
		PFD_TYPE_RGBA,                   // RGBA color mode
		32,                              // 32 bit color mode
		0, 0, 0, 0, 0, 0,                // ignore color bits, non-palettized mode
		0,                               // no alpha buffer
		0,                               // ignore shift bit
		0,                               // no accumulation buffer
		0, 0, 0, 0,                      // ignore accumulation bits
		16,                              // 16 bit z-buffer size
		0,                               // no stencil buffer
		0,                               // no auxiliary buffer
		PFD_MAIN_PLANE,                  // main drawing plane
		0,                               // reserved
		0, 0, 0 };                       // layer masks ignored

	nPixelFormat = ChoosePixelFormat(hDC, &pfd); // choose best matching pixel format

	SetPixelFormat(hDC, nPixelFormat, &pfd);     // set pixel format to device context
	}

//////////////////////////////////////////////////////////////////////////////// GL_CalculateNormal

// desc: given 3 points, calculates the normal to the points
void GL_CalculateNormal( double *p1, double *p2, double *p3 )
	{
	double a[3], b[3], result[3];
	double length;

	a[0] = p1[0] - p2[0];
	a[1] = p1[1] - p2[1];
	a[2] = p1[2] - p2[2];

	b[0] = p1[0] - p3[0];
	b[1] = p1[1] - p3[1];
	b[2] = p1[2] - p3[2];

	result[0] = a[1] * b[2] - b[1] * a[2];
	result[1] = b[0] * a[2] - a[0] * b[2];
	result[2] = a[0] * b[1] - b[0] * a[1];

	// calculate the length of the normal
	length = (double)sqrt(result[0]*result[0] + result[1]*result[1] + result[2]*result[2]);

	// normalize and specify the normal
	glNormal3d(result[0]/length, result[1]/length, result[2]/length);
	}

void GL_InitializeOrthoView()
	{
	double dim = MAXVAL * 1.25;

	glViewport(0, 0, grectGLV.right, grectGLV.bottom);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	if (gbModelMode_Active)
		{
		if (grectGLV.bottom == grectGLV.right)
			glOrtho(-dim, dim, -dim, dim, 0.0, MAXSCALE * MAXVAL * 2);
		else if (grectGLV.bottom > grectGLV.right)
			glOrtho(-(double)grectGLV.right/grectGLV.bottom*dim, (double)grectGLV.right/grectGLV.bottom*dim, -dim, dim, 0.0, MAXSCALE * MAXVAL * 2);
		else
			glOrtho(-dim, dim, -(double)grectGLV.bottom/grectGLV.right*dim, (double)grectGLV.bottom/grectGLV.right*dim, 0.0, MAXSCALE * MAXVAL * 2);
		}
/*
	else if (gbUVMapMode_Active)
		{
		glOrtho(guvmap->minS-guvmap->avgS, guvmap->maxS-guvmap->avgS, guvmap->minT-guvmap->avgT, guvmap->maxT-guvmap->avgT, -10.0, 10.0);
		//glOrtho(-0.5, 0.5, -0.5, 0.5, -10.0, 10.0);
		}
*/
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	}

void GL_ClearGLV()
	{
	if (ghdcGLV)
		{
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glFlush();
		SwapBuffers(ghdcGLV);
		}
	}

/////////////////////////////////////////////////////////////////////////////////// WndProc_Options

LRESULT CALLBACK WndProc_Options(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
	switch (uMsg)
		{
		case WM_KEYDOWN:
			{
			switch (wParam)
				{
				case VK_ESCAPE:
					{
					SendMessage(ghwndOptions, WM_CLOSE, 0, 0);
					break;
					}
				case VK_RETURN:
					{
					SendMessage(ghwndOptions, WM_COMMAND, (WPARAM)ID_OPTIONSOK, 0);
					break;
					}
				default:
					break;
				}
			break;
			}

		case WM_UPDATE:
			{
			if (gBmpFirst)
				CheckRadioButton(hWnd, ID_OPTIONSMATFIRST, ID_OPTIONSBMPFIRST, ID_OPTIONSBMPFIRST);
			else
				CheckRadioButton(hWnd, ID_OPTIONSMATFIRST, ID_OPTIONSBMPFIRST, ID_OPTIONSMATFIRST);

			if (gMatSecond)
				CheckDlgButton(hWnd, ID_OPTIONSMATSECOND, 1);
			else
				CheckDlgButton(hWnd, ID_OPTIONSMATSECOND, 0);
			if (gBmpSecond)
				CheckDlgButton(hWnd, ID_OPTIONSBMPSECOND, 1);
			else
				CheckDlgButton(hWnd, ID_OPTIONSBMPSECOND, 0);

			SetWindowText(ghwndOptionsCmpFile2, (char*)gCmpPath);

			break;
			}

		case WM_CLOSE:
			{
			ShowWindow(ghwndOptions, SW_HIDE);
			gbOptionsShow = false;
			break;
			}

		case WM_USERCLOSE:
			{
			DestroyWindow(ghwndOptions);
			break;
			}

		case WM_COMMAND:
			{
			switch(LOWORD(wParam))
				{
				case ID_OPTIONSMATFIRST:
					{
					CheckRadioButton(hWnd, ID_OPTIONSMATFIRST, ID_OPTIONSBMPFIRST, ID_OPTIONSMATFIRST);
					break;
					} // end case LOWORD = ID_OPTIONSMATFIRST

				case ID_OPTIONSBMPFIRST:
					{
					CheckRadioButton(hWnd, ID_OPTIONSMATFIRST, ID_OPTIONSBMPFIRST, ID_OPTIONSBMPFIRST);
					break;
					} // end case LOWORD = ID_OPTIONSBMPFIRST

				case ID_OPTIONSCMPBROWSE:
					{
					static char szPathname  [_MAX_PATH];
					static char szTitlename [_MAX_FNAME];
					static char szFilterCmp [] = "CMP palette\0*.cmp\0\0";

					if (FileDlg_GetOpenPathname(ghwndOptions, szPathname, szTitlename, szFilterCmp))
						SetWindowText(ghwndOptionsCmpFile2, szPathname);

					break;
					} // end case LOWORD = ID_GROUPCHANGEBMP

				case ID_OPTIONSCMPDEFAULT:
					{
					SetWindowText(ghwndOptionsCmpFile2, (char*)gszDfltCmp);
					break;
					}

				case ID_OPTIONSOK:
					{
					if (IsDlgButtonChecked(hWnd, ID_OPTIONSBMPFIRST))
						{
						gMatFirst = false;
						gBmpFirst = true;
						}
					else
						{
						gMatFirst = true;
						gBmpFirst = false;
						}

					if (!IsDlgButtonChecked(hWnd, ID_OPTIONSMATSECOND))
						gMatSecond = false;
					else
						gMatSecond = true;

					if (!IsDlgButtonChecked(hWnd, ID_OPTIONSBMPSECOND))
						gBmpSecond = false;
					else
						gBmpSecond = true;

					GetWindowText(ghwndOptionsCmpFile2, (char*)gCmpPath, _MAX_PATH);

					Options_SaveToReg();

					LoadCmpFile();

					if (gmodel)
						gmodel->Reload8bitTextures();

					SendMessage(ghwndOptions, WM_CLOSE, 0, 0);
					break;
					} // end case LOWORD = ID_OPTIONSOK

				case ID_OPTIONSCANCEL:
					{
					SendMessage(ghwndOptions, WM_CLOSE, 0, 0);
					break;
					} // end case LOWORD = ID_OPTIONSCANCEL

				default:
					break;
				} // end switch (LOWORD)

			} // end case uMsg = WM_COMMAND

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
		} // end switch (uMsg)

	return 0;
	}


//////////////////////////////////////////////////////////////////////////////// WndProc_GLViewport

LRESULT CALLBACK WndProc_GLViewport(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
	int oldMouseX, oldMouseY;

	static double xpos, ypos;
	static double oldxpos, oldypos;

	static bool m_down  = false;
	static bool m_moved = false;

	switch (uMsg)
		{
		case WM_SIZE:
			{
			GetClientRect(ghwndGLV, &grectGLV);

			GL_InitializeOrthoView(); // re-calculate the projection
			if(gbModelMode_Active)
				{
				gCamera->doLookAt();
				gmodel->Display();
				}
			else if(gbUVMapMode_Active)
				{
				guvmap->SetupOrthoView();
				guvmap->Display();
				}

			break;
			}

		case WM_MOUSEMOVE:
			{
			if ((gbModelMode_Active) && (gCamera->getMouseLook()))
				{
				// save old mouse coordinates
				oldMouseX = giMouseX;
				oldMouseY = giMouseY;

				// get new mouse coordinates from Windows
				giMouseX = LOWORD(lParam);
				giMouseY = HIWORD(lParam);

				// mouse moved to the right
				if ((giMouseX - oldMouseX) > 0)
					gCamera->AngleH(1, false);
				// mouse moved to the left
				else if ((giMouseX - oldMouseX) < 0)
					gCamera->AngleH(-1, false);

				// mouse moved up
				if ((giMouseY - oldMouseY) < 0)
					gCamera->AngleV(1, false);
				// mouse moved down
				else if ((giMouseY - oldMouseY) > 0)
					gCamera->AngleV(-1, false);
				}

			else if (gbUVMapMode_Active)
				{
				oldxpos = xpos;
				oldypos = ypos;

				xpos =       (double)LOWORD(lParam)/grectGLV.right;
				ypos = 1.0 - (double)HIWORD(lParam)/grectGLV.bottom;

				if (! m_moved)
					{
					m_moved = true;
					guvmap->Select_RectStart(oldxpos, oldypos);
					}

				if (gbKeyPressed[VK_CONTROL])
					guvmap->MoveSelection(xpos-oldxpos, ypos-oldypos);
				else if(m_down)
					guvmap->Select_RectSize(xpos-oldxpos, ypos-oldypos);
				}

			break;
			}

		case WM_LBUTTONDOWN:
			{
			if (gbUVMapMode_Active)
				{
				m_down  = true;
				m_moved = false;
				}

			break;
			}

		case WM_LBUTTONUP:
			{
			if (gbUVMapMode_Active && m_down)
				{
				if (m_moved)
					guvmap->Select_RectEnd();
				else
					guvmap->Select_Point(xpos, ypos);

				m_down  = false;
				m_moved = false;
				}

			break;
			}

		/*case WM_NCLBUTTONUP:
			{
			if (gbUVMapMode_Active)
				guvmap->Select_RectEnd();
			break;
			}

		case WM_CAPTURECHANGED:
			{
			MessageBox(NULL, "", "", NULL);
			break;
			}*/

		case WM_CLOSE:
			{
			// deselect and delete rendering context, release device context
			wglMakeCurrent(NULL, NULL);
			wglDeleteContext(ghrcGLV);
			ReleaseDC(ghwndGLV, ghdcGLV);

			DestroyWindow(ghwndGLV);

			break;
			}

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
		}

	return 0;
	}

//////////////////////////////////////////////////////////////////////////////// WndProc_Controls

LRESULT CALLBACK WndProc_Controls(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
	static char szPathname  [_MAX_PATH];
	static char szTitlename [_MAX_FNAME];
	static char szFilterBoth[] = "MAT and BMP\0*.mat;*.bmp\0MAT\0*.mat\0BMP\0*.bmp\0\0";

	int i, sel, sel2;

	switch (uMsg)
		{
		case WM_CLOSE:
			{
			ShowWindow(ghwndControls, SW_HIDE);
			gbControlPanelShow = false;
			break;
			}

		case WM_USERCLOSE:
			{
			DestroyWindow(ghwndControls);
			break;
			}

		case WM_COMMAND:
			{
			switch(LOWORD(wParam))
				{
				case ID_MESHLB:
					{
					switch (HIWORD(wParam))
						{
						case LBN_SELCHANGE:
							{
							if((sel = SendMessage(ghwndMeshLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
								{
								SendMessage(ghwndGroupLB, LB_RESETCONTENT, 0, 0);
								for(i = 0; i < gmodel->meshes[sel].numglGroups; i++)
									SendMessage(ghwndGroupLB, LB_INSERTSTRING, i, (LPARAM)gmodel->meshes[sel].glGroups[i].glGroupName);
								PostMessage(ghwndGroupLB, WM_PAINT, 0, 0);

								gmodel->Deselect_AllMeshes();

								if (giProgramMode == 3)
									{
									ProgramMode2to3(false);
									if (guvmap != NULL)
										delete guvmap;
									guvmap = NULL;
									}
								}
							break;
							} // end case HIWORD = LBN_SELCHANGE

						case LBN_DBLCLK:
							{
							if((sel = SendMessage(ghwndMeshLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
								{
								if (gmodel->meshes[sel].glRenderMode == HIDE)
									gmodel->meshes[sel].glRenderMode = WIRE;
								else if (gmodel->meshes[sel].glRenderMode == WIRE)
									gmodel->meshes[sel].glRenderMode = SHOW;
								else if (gmodel->meshes[sel].glRenderMode == SHOW)
									gmodel->meshes[sel].glRenderMode = HIDE;
								}
							break;
							} // end case HIWORD = LBN_DBLCLK

						default:
							break;
						} // end switch (HIWORD)

					break;
					} // end case LOWORD = ID_MESHLB

				case ID_MESHHIDE:
					{
					SetFocus(ghwndMeshLB);
					if((sel = SendMessage(ghwndMeshLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
						gmodel->meshes[sel].glRenderMode = HIDE;
					break;
					} // end case LOWORD = ID_MESHHIDE

				case ID_MESHWIRE:
					{
					SetFocus(ghwndMeshLB);
					if((sel = SendMessage(ghwndMeshLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
						gmodel->meshes[sel].glRenderMode = WIRE;
					break;
					} // end case LOWORD = ID_MESHWIRE

				case ID_MESHSHOW:
					{
					SetFocus(ghwndMeshLB);
					if((sel = SendMessage(ghwndMeshLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
						gmodel->meshes[sel].glRenderMode = SHOW;
					break;
					} // end case LOWORD = ID_MESHSHOW

				case ID_MESHHIDEALL:
					{
					SetFocus(ghwndMeshLB);
					for(i = 0; i < gmodel->numMeshes; i++)
						gmodel->meshes[i].glRenderMode = HIDE;
					break;
					} // end case LOWORD = ID_MESHHIDEALL

				case ID_MESHWIREALL:
					{
					SetFocus(ghwndMeshLB);
					for(i = 0; i < gmodel->numMeshes; i++)
						gmodel->meshes[i].glRenderMode = WIRE;
					break;
					} // end case LOWORD = ID_MESHWIREALL

				case ID_MESHSHOWALL:
					{
					SetFocus(ghwndMeshLB);
					for(i = 0; i < gmodel->numMeshes; i++)
						gmodel->meshes[i].glRenderMode = SHOW;
					break;
					} // end case LOWORD = ID_MESHSHOWALL

				case ID_GROUPLB:
					{
					switch (HIWORD(wParam))
						{
						case LBN_SELCHANGE:
							{
							gmodel->Deselect_AllMeshes();

							if((sel = SendMessage(ghwndMeshLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
								if((sel2 = SendMessage(ghwndGroupLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
									{
									gmodel->meshes[sel].Select_Group(sel2);

									if (giProgramMode == 2)
										ProgramMode2to3(true);

									if (guvmap != NULL)
										delete guvmap;
									guvmap = new CUVMap(sel, sel2);
									}
							break;
							} // end case HIWORD = LBN_SELCHANGE

						case LBN_DBLCLK:
							{
							if((sel = SendMessage(ghwndMeshLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
								if((sel2 = SendMessage(ghwndGroupLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
									{
									strcpy(szPathname,  gszEmptyString);
									strcpy(szTitlename, gszEmptyString);
									if (FileDlg_GetOpenPathname(ghwndControls, szPathname, szTitlename, szFilterBoth))
										{
										gmodel->Change_Tex(sel, sel2, szTitlename, szPathname);

										// update the group list box
										SendMessage(ghwndGroupLB, LB_RESETCONTENT, 0, 0);
										for(i = 0; i < gmodel->meshes[sel].numglGroups; i++)
											SendMessage(ghwndGroupLB, LB_INSERTSTRING, i, (LPARAM)gmodel->meshes[sel].glGroups[i].glGroupName);
										PostMessage(ghwndGroupLB, WM_PAINT, 0, 0);

										// clear the selection
										gmodel->meshes[sel].Deselect_AllGroups();

										// reselect the group that was changed
										SendMessage(ghwndGroupLB, LB_SETCURSEL, sel2, 0);
										gmodel->meshes[sel].Select_Group(sel2);

										if (guvmap != NULL)
											delete guvmap;
										guvmap = new CUVMap(sel, sel2);
										}
									}
							SetFocus(ghwndGroupLB);
							break;
							} // end case HIWORD = LBN_DBLCLK

						default:
							break;
						} // end switch (HIWORD)

					break;
					} // end case LOWORD = ID_GROUPLB

				case ID_GROUPSELECTNONE:
					{
					SetFocus(ghwndMeshLB);

					// set list box to have no selection
					SendMessage(ghwndGroupLB, LB_SETCURSEL, -1, 0);
					PostMessage(ghwndGroupLB, WM_PAINT, 0, 0);

					// deselect all groups in the model
					gmodel->Deselect_AllMeshes();

					ProgramMode2to3(false);

					if (guvmap != NULL)
						delete guvmap;
					guvmap = NULL;

					break;
					} // end case LOWORD = ID_GROUPSELECTNONE

				case ID_GROUPCHANGEBMP:
					{
					if((sel = SendMessage(ghwndMeshLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
						if((sel2 = SendMessage(ghwndGroupLB, LB_GETCURSEL, 0, 0)) != LB_ERR)
							{
							strcpy(szPathname,  gszEmptyString);
							strcpy(szTitlename, gszEmptyString);
							if (FileDlg_GetOpenPathname(ghwndControls, szPathname, szTitlename, szFilterBoth))
								{
								gmodel->Change_Tex(sel, sel2, szTitlename, szPathname);

								// update the group list box
								SendMessage(ghwndGroupLB, LB_RESETCONTENT, 0, 0);
								for(i = 0; i < gmodel->meshes[sel].numglGroups; i++)
									SendMessage(ghwndGroupLB, LB_INSERTSTRING, i, (LPARAM)gmodel->meshes[sel].glGroups[i].glGroupName);
								PostMessage(ghwndGroupLB, WM_PAINT, 0, 0);

								// clear the selection
								gmodel->meshes[sel].Deselect_AllGroups();

								// reselect the group that was changed
								SendMessage(ghwndGroupLB, LB_SETCURSEL, sel2, 0);
								gmodel->meshes[sel].Select_Group(sel2);

								if (guvmap != NULL)
									delete guvmap;
								guvmap = new CUVMap(sel, sel2);
								}
							}
					SetFocus(ghwndGroupLB);
					break;
					} // end case LOWORD = ID_GROUPCHANGEBMP

				case ID_UVORTHOFRONT:
					{
					guvmap->Remap(ID_UVORTHOFRONT);
					SetFocus(ghwndControls);
					break;
					} // end case LOWORD = ID_UVORTHOFRONT

				case ID_UVORTHOBACK:
					{
					guvmap->Remap(ID_UVORTHOBACK);
					SetFocus(ghwndControls);
					break;
					} // end case LOWORD = ID_UVORTHOBACK

				case ID_UVORTHOTOP:
					{
					guvmap->Remap(ID_UVORTHOTOP);
					SetFocus(ghwndControls);
					break;
					} // end case LOWORD = ID_UVORTHOTOP

				case ID_UVORTHOBOTTOM:
					{
					guvmap->Remap(ID_UVORTHOBOTTOM);
					SetFocus(ghwndControls);
					break;
					} // end case LOWORD = ID_UVORTHOBOTTOM

				case ID_UVORTHOLEFT:
					{
					guvmap->Remap(ID_UVORTHOLEFT);
					SetFocus(ghwndControls);
					break;
					} // end case LOWORD = ID_UVORTHOLEFT

				case ID_UVORTHORIGHT:
					{
					guvmap->Remap(ID_UVORTHORIGHT);
					SetFocus(ghwndControls);
					break;
					} // end case LOWORD = ID_UVORTHORIGHT

				case ID_UVEDIT:
					{
					if (gbModelMode_Active)
						{
						ProgramMode3to4(true);
						gbModelMode_Active = false;
						gbUVMapMode_Active = true;

						guvmap->SetupOrthoView();
						guvmap->Display();
						}
					else if (gbUVMapMode_Active)
						{
						ProgramMode3to4(false);
						gbModelMode_Active = true;
						gbUVMapMode_Active = false;

						GL_InitializeOrthoView();
						GL_RenderScene();
						}
					SetFocus(ghwndControls);
					break;
					} // end case LOWORD = ID_UVEDIT

				case ID_UVHELP:
					{
					MessageBox(ghwndControls, "Drag a rectangle or click on individual vertices to make a selection.\n\nHold CTRL and use the mouse to move the selection.\n\nScaling Tips:\n\n     * Scale by -1 to flip the vertices around the axis.\n\n     * Scale by 0 to ''merge'' vertices.\n\n     * Scale up beyond the edges of the texture to create a tiling effect.\n       (This can be put to good use in level design)", gszAppName, MB_OK | MB_APPLMODAL);
					SetFocus(ghwndControls);
					break;
					} // end case LOWORD = ID_UVHELP

				case ID_UVSELECTALL:
					{
					SetFocus(ghwndControls);
					guvmap->Select_All();
					break;
					} // end case LOWORD = ID_UVSELECTALL

				case ID_UVSELECTNONE:
					{
					SetFocus(ghwndControls);
					guvmap->Select_None();
					break;
					} // end case LOWORD = ID_UVSELECTNONE

				case ID_UVSELECTINVERT:
					{
					SetFocus(ghwndControls);
					guvmap->Select_Invert();
					break;
					} // end case LOWORD = ID_UVSELECTINVERT

				case ID_UVSCALE:
					{
					char ScaleString[16];
					double ScaleAmount;

					if (guvmap->TestForSelection())
						if (GetWindowText(ghwndUVScaleAmount, ScaleString, 16) > 0)
							{
							ScaleAmount = atof(ScaleString);
								if (BST_CHECKED == SendMessage(ghwndUVScaleX, BM_GETCHECK, 0, 0))
									guvmap->ScaleX(ScaleAmount);
								else if (BST_CHECKED == SendMessage(ghwndUVScaleY, BM_GETCHECK, 0, 0))
									guvmap->ScaleY(ScaleAmount);
								else if (BST_CHECKED == SendMessage(ghwndUVScaleBoth, BM_GETCHECK, 0, 0))
									{
									guvmap->ScaleX(ScaleAmount);
									guvmap->ScaleY(ScaleAmount);
									}
							}

					SetFocus(ghwndControls);
					break;
					} // end case LOWORD = ID_UVSCALE

				case ID_UVXAXIS:
					{
					CheckRadioButton(hWnd, ID_UVXAXIS, ID_UVBOTHAXES, ID_UVXAXIS);
					SetFocus(ghwndControls);
					break;
					}
				case ID_UVYAXIS:
					{
					CheckRadioButton(hWnd, ID_UVXAXIS, ID_UVBOTHAXES, ID_UVYAXIS);
					SetFocus(ghwndControls);
					break;
					}
				case ID_UVBOTHAXES:
					{
					CheckRadioButton(hWnd, ID_UVXAXIS, ID_UVBOTHAXES, ID_UVBOTHAXES);
					SetFocus(ghwndControls);
					break;
					}

				default:
					break;
				} // end switch (LOWORD)

			} // end case uMsg = WM_COMMAND

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
		} // end switch (uMsg)

	return 0;
	}

////////////////////////////////////////////////////////////////////////////////////// WndProc_Main

LRESULT CALLBACK WndProc_Main(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
	static char szPathname  [_MAX_PATH];
	static char szTitlename [_MAX_FNAME];
	static char szFilter3do [] = "3DO model\0*.3do\0\0";
	char extension[4];

	int cxClient, cyClient;

	HWND hwndScrolling;

	/////////////////////

	switch (uMsg)
		{
		case WM_SIZE:
			{
			cxClient = LOWORD(lParam);
			cyClient = HIWORD(lParam);

			MoveWindow(ghwndGLV, 2*SEP+BUTTON, SEP, cxClient-4*SEP-2*BUTTON, cyClient-3*SEP-BUTTON, true);

			MoveWindow(ghwndScrollPanY,  SEP, SEP, SB, cyClient-2*SEP-BUTTON, true);
			MoveWindow(ghwndScrollPitch, SEP+SB+1, SEP, SB, cyClient-2*SEP-BUTTON, true);

			MoveWindow(ghwndScrollYaw,   SEP+BUTTON, cyClient-SEP-BUTTON, cxClient-2*SEP-2*BUTTON-1, SB, true);
			MoveWindow(ghwndScrollPanX,  SEP+BUTTON, cyClient-SEP-SB+2, cxClient-2*SEP-2*BUTTON-1, SB, true);

			MoveWindow(ghwndScrollZoom,  cxClient-SEP-BUTTON, SEP, BUTTON, cyClient-2*SEP-BUTTON, true);

			MoveWindow(ghwndMouseLook,   SEP, cyClient-SEP-BUTTON, BUTTON, BUTTON, true);
			MoveWindow(ghwndResetView,   cxClient-SEP-BUTTON, cyClient-SEP-BUTTON, BUTTON, BUTTON, true);

			break;
			}

		case WM_VSCROLL:
		case WM_HSCROLL:
			{
			hwndScrolling = (HWND)lParam;

			switch(LOWORD(wParam))
				{
				case SB_LINEDOWN:
					{
					if (hwndScrolling == ghwndScrollPitch)
						gCamera->AngleV(-1, false);

					else if (hwndScrolling == ghwndScrollYaw)
						gCamera->AngleH(1, false);

					else if (hwndScrolling == ghwndScrollZoom)
						gCamera->Zoom(-1);

					else if (hwndScrolling == ghwndScrollPanY)
						gCamera->PanY(1);

					else if (hwndScrolling == ghwndScrollPanX)
						gCamera->PanX(1);

					break;
					}

				case SB_LINEUP:
					{
					if (hwndScrolling == ghwndScrollPitch)
						gCamera->AngleV(1, false);

					else if (hwndScrolling == ghwndScrollYaw)
						gCamera->AngleH(-1, false);

					else if (hwndScrolling == ghwndScrollZoom)
						gCamera->Zoom(1);

					else if (hwndScrolling == ghwndScrollPanY)
						gCamera->PanY(-1);

					else if (hwndScrolling == ghwndScrollPanX)
						gCamera->PanX(-1);

					break;
					}

				case SB_PAGEDOWN:
					{
					if (hwndScrolling == ghwndScrollPitch)
						gCamera->AngleV(-10, true);

					else if (hwndScrolling == ghwndScrollYaw)
						gCamera->AngleH(10, true);

					else if (hwndScrolling == ghwndScrollZoom)
						gCamera->Zoom(-10);

					else if (hwndScrolling == ghwndScrollPanY)
						gCamera->PanY(10);

					else if (hwndScrolling == ghwndScrollPanX)
						gCamera->PanX(10);

					break;
					}

				case SB_PAGEUP:
					{
					if (hwndScrolling == ghwndScrollPitch)
						gCamera->AngleV(10, true);

					else if (hwndScrolling == ghwndScrollYaw)
						gCamera->AngleH(-10, true);

					else if (hwndScrolling == ghwndScrollZoom)
						gCamera->Zoom(10);

					else if (hwndScrolling == ghwndScrollPanY)
						gCamera->PanY(-10);

					else if (hwndScrolling == ghwndScrollPanX)
						gCamera->PanX(-10);

					break;
					}

				case SB_THUMBTRACK:
				case SB_THUMBPOSITION:
					{
					if (hwndScrolling == ghwndScrollPitch)
						gCamera->AngleV(-(short int)HIWORD(wParam) - (int)gCamera->getVAngle(), true);

					else if (hwndScrolling == ghwndScrollYaw)
						gCamera->AngleH((short int)HIWORD(wParam) - (int)gCamera->getHAngle(), true);

					else if (hwndScrolling == ghwndScrollZoom)
						gCamera->Zoom(-(short int)HIWORD(wParam) - gCamera->getZoompos());

					else if (hwndScrolling == ghwndScrollPanY)
						gCamera->PanY((short int)HIWORD(wParam) - gCamera->getYscrollpos());

					else if (hwndScrolling == ghwndScrollPanX)
						gCamera->PanX((short int)HIWORD(wParam) - gCamera->getXscrollpos());

					break;
					}

				default:
					break;
				}

			if (gbModelMode_Active)
				GL_RenderScene();

			break;
			}

		case WM_COMMAND:
			{
			switch (LOWORD(wParam))
				{
				case ID_MENUOFFSET:
					{
					if (gmodel != NULL)
						{
						char message[1024];
						sprintf(message, "This will adjust the insert offset so that the\ncharacter's feet are aligned with the ground in-game.\n\nThis is not desirable for non-character 3do's.\n\nIf you choose Yes, the offset will be adjusted to:\n0   0   %f", gmodel->glMinZ);
						if (IDYES == MessageBox(ghwndMain, message, "Re-calculate character offset", MB_YESNO | MB_APPLMODAL))
							{
							gmodel->jkInsertOffset.x = 0;
							gmodel->jkInsertOffset.y = 0;
							gmodel->jkInsertOffset.z = gmodel->glMinZ;
							}
						}
					else
						{
						MessageBox(ghwndMain, "You must open a 3do file first.", "Re-calculate character offset", MB_APPLMODAL);
						}
					break;
					}

				case ID_MOUSELOOK:
					{
					SetFocus(ghwndMain);
					gCamera->ToggleMouseLook();
					break;
					}

				case ID_RESETVIEW:
					{
					SetFocus(ghwndMain);
					gCamera->ResetView(gmodel->radius);
					break;
					}

				case ID_MENUOPTIONS:
					{
					SendMessage(ghwndOptions, WM_UPDATE, 0, 0);
					ShowWindow(ghwndOptions, SW_NORMAL);
					gbOptionsShow = true;
					SetFocus(ghwndOptions);
					break;
					}

				case ID_MENUHIDECONTROLS:
					{
					if (!gbControlPanelShow)
						ShowWindow(ghwndControls, SW_NORMAL);
					else
						ShowWindow(ghwndControls, SW_HIDE);

					gbControlPanelShow = !gbControlPanelShow;
					break;
					}

				case ID_MENUOPEN:
					{
					if (gmodel != NULL)
						{
						if (IDYES == MessageBox(ghwndMain, "Save?", gszAppName, MB_YESNO | MB_APPLMODAL))
							{
							strcpy(szPathname,  gszProjectPathname);
							strcpy(szTitlename, gszProjectTitlename);
							if (FileDlg_GetSavePathname(ghwndMain, szPathname, szTitlename, szFilter3do))
								Command_SaveAs(szPathname);
							}
						}

					strcpy(szPathname,  gszEmptyString);
					strcpy(szTitlename, gszEmptyString);
					if (FileDlg_GetOpenPathname(ghwndMain, szPathname, szTitlename, szFilter3do))
						{
						extension[0] = szTitlename[strlen(szTitlename) -3];
						extension[1] = szTitlename[strlen(szTitlename) -2];
						extension[2] = szTitlename[strlen(szTitlename) -1];
						extension[3] = '\0';

						if(!(strcmpi(extension, "3do")))
							{
							if (gmodel != NULL)
								Command_Close();
							Command_Open(szPathname);
							}
						else
							MessageBox(ghwndMain, "Please select a .3do file.", gszAppName, MB_APPLMODAL);
						}
					break;
					} // end case LOWORD = ID_MENUOPEN

				case ID_MENUREVERT:
					{
					if (IDYES == MessageBox(ghwndMain, "This will revert to the saved file.\n\nAre you sure?", gszAppName, MB_YESNO | MB_APPLMODAL))
						{
						strcpy(szPathname, gszProjectPathname);
						Command_Close();
						Command_Open(szPathname);
						}
					break;
					} // end case LOWORD = ID_MENUREVERT

				case ID_MENUCLOSE:
					{
					if (IDYES == MessageBox(ghwndMain, "Save?", gszAppName, MB_YESNO | MB_APPLMODAL))
						{
						strcpy(szPathname,  gszProjectPathname);
						strcpy(szTitlename, gszProjectTitlename);
						if (FileDlg_GetSavePathname(ghwndMain, szPathname, szTitlename, szFilter3do))
							Command_SaveAs(szPathname);
						}

					Command_Close();
					break;
					} // end case LOWORD = ID_MENUCLOSE

				case ID_MENUSAVE:
					{
					Command_Save();
					break;
					} // end case LOWORD = ID_MENUSAVE

				case ID_MENUSAVEAS:
					{
					strcpy(szPathname,  gszProjectPathname);
					strcpy(szTitlename, gszProjectTitlename);
					if (FileDlg_GetSavePathname(ghwndMain, szPathname, szTitlename, szFilter3do))
						Command_SaveAs(szPathname);
					break;
					} // end case LOWORD = ID_MENUSAVEAS

				case ID_MENUABOUT:
					{
					MessageBox(ghwndMain, "by Geoff Archer - garcher@alum.wpi.edu\n\nhttp://www.thecia.net/users/garcher/forceremap/\n\nThis program is freeware and should be distributed as such.", gszAppName, MB_APPLMODAL);
					break;
					} // end case LOWORD = ID_MENUABOUT

				default:
					break;

				} // end switch (LOWORD)

			break;
			} // end case uMsg = WM_COMMAND

		case WM_CHAR:
			{
			switch (toupper(wParam))
				{
				case 'M':
					{
					if(gbModelMode_Active)
						if(gmodel)
							gCamera->ToggleMouseLook();
					break;
					}

				case 'R':
					{
					if(gbModelMode_Active)
						if(gmodel)
							gCamera->ResetView(gmodel->radius);
					break;
					}

				default:
					break;
				} // end switch (WPARAM)
			break;
			} // end case uMsg = WM_CHAR

		case WM_KEYDOWN:
			{
			gbKeyPressed[wParam] = true;
			//Msg("", "%d", wParam);

			switch (wParam)
				{
				case VK_UP:
					{
					if(gbModelMode_Active)
						if(gmodel)
							{
							if(gbKeyPressed[VK_SHIFT])
								if (gbKeyPressed[17])
									gCamera->AngleV(10, false);
								else
									gCamera->AngleV(1, false);

							else if(gbKeyPressed[17])
								gCamera->PanY(-10);

							else
								gCamera->PanY(-1);
							}
					break;
					}

				case VK_DOWN:
					{
					if(gbModelMode_Active)
						if(gmodel)
							{
							if(gbKeyPressed[VK_SHIFT])
								if (gbKeyPressed[17])
									gCamera->AngleV(-10, false);
								else
									gCamera->AngleV(-1, false);

							else if(gbKeyPressed[17])
								gCamera->PanY(10);

							else
								gCamera->PanY(1);
							}
					break;
					}

				case VK_LEFT:
					{
					if(gbModelMode_Active)
						if(gmodel)
							{
							if(gbKeyPressed[VK_SHIFT])
								if (gbKeyPressed[17])
									gCamera->AngleH(-10, false);
								else
									gCamera->AngleH(-1, false);

							else if(gbKeyPressed[17])
								gCamera->PanX(-10);

							else
								gCamera->PanX(-1);
							}
					break;
					}

				case VK_RIGHT:
					{
					if(gbModelMode_Active)
						if(gmodel)
							{
							if(gbKeyPressed[VK_SHIFT])
								if (gbKeyPressed[17])
									gCamera->AngleH(10, false);
								else
									gCamera->AngleH(1, false);

							else if(gbKeyPressed[17])
								gCamera->PanX(10);

							else
								gCamera->PanX(1);
							}
					break;
					}

				case 107: // NUMPAD +
					{
					if(gbModelMode_Active)
						if(gmodel)
							{
							if (gbKeyPressed[17])
								gCamera->Zoom(10);
							else
								gCamera->Zoom(1);
							}
					break;
					}

				case 109: // NUMPAD -
					{
					if(gbModelMode_Active)
						if(gmodel)
							{
							if (gbKeyPressed[17])
								gCamera->Zoom(-10);
							else
								gCamera->Zoom(-1);
							}
					break;
					}

				case 106: // NUMPAD *
					{
					if(gbModelMode_Active)
						if(gmodel)
							{
							if (gbKeyPressed[17])
								{
								gCamera->AngleV(5, false);
								gCamera->AngleH(10, false);
								}
							else
								{
								gCamera->AngleV(1, false);
								gCamera->AngleH(2, false);
								}
							}
					break;
					}

				case 8: // BACKSPACE: select none
					{
					if(gbModelMode_Active)
						{
						if(SendMessage(ghwndMeshLB, LB_GETCURSEL, 0, 0) != LB_ERR)
							{
							// set list box to have no selection
							SendMessage(ghwndGroupLB, LB_SETCURSEL, -1, 0);
							PostMessage(ghwndGroupLB, WM_PAINT, 0, 0);

							// deselect all groups in the model
							gmodel->Deselect_AllMeshes();

							ProgramMode2to3(false);
							}
						}
					else if (gbUVMapMode_Active)
						{
						guvmap->Select_None();
						}
					break;
					} // end case LOWORD = ID_GROUPSELECTNONE

				case VK_RETURN:
					{
					if (!gbControlPanelShow)
						ShowWindow(ghwndControls, SW_NORMAL);
					else
						ShowWindow(ghwndControls, SW_HIDE);

					gbControlPanelShow = !gbControlPanelShow;
					SetFocus(ghwndMain);
					break;
					}

				default:
					break;
				} // end switch (WPARAM)

			break;
			} // end case uMsg = WM_KEYDOWN

		case WM_KEYUP:
			{
			gbKeyPressed[wParam] = false;

			break;
			} // end case uMsg = WM_KEYUP

		case WM_CLOSE:
			{
			gbModelMode_Active = false;
			gbUVMapMode_Active = false;

			if (gmodel != NULL)
				delete gmodel;
			gmodel = NULL;

			if (guvmap != NULL)
				delete guvmap;
			guvmap = NULL;

			ReleaseCapture();

			AllWindows_Close();

			/*DestroyWindow(ghwndMain);*/

			PostQuitMessage(0);
			break;
			} // end case uMsg = WM_CLOSE

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
		} // end switch (uMsg)

	return 0;
	}

/////////////////////////////////////////////////////////////////////////////////////////// WinMain

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
	{
	MSG   msg;
	bool  fDone;

	// set the global application instance
	ghinstApp = hInstance;

	// init global objects
	gCamera = new Camera;
	gmodel  = NULL;
	guvmap  = NULL;
	gszEmptyString[0] = '\0';
	Support_SetGlobalFilenameInfo(gszEmptyString);
	Options_LoadFromReg();

	// initialize program flags
	fDone = false;
	gbModelMode_Active = false;
	gbUVMapMode_Active = false;
	gCmpHasBeenLoaded = false;

	// create all child windows
	if (!AllWindows_Create())
		{
		MessageBox(NULL, "Window creation failed.", "Windows error", MB_ICONEXCLAMATION | MB_APPLMODAL);
		return false;
		}

	ProgramMode1();
	gbControlPanelShow = true;
	gbOptionsShow = false;
	FileDlg_Init();

	// main message loop
	while((GetMessage (&msg, NULL, 0, 0)) && (!fDone))
		{
		if (msg.message == WM_QUIT)
			{
			fDone = true;
			}
		else
			{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			}

		GL_RenderScene();
		}

	/*MessageBox(NULL, "wait", "wait", NULL);*/

	return msg.wParam;
	}

///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////// AllWindows_Create

bool AllWindows_Create()
	{
	WNDCLASSEX  wc;

	HMENU       hmenuPopupFile;
	HMENU       hmenuPopupOptions;

	SCROLLINFO  si;

	DWORD       dwMainStyle = WS_OVERLAPPEDWINDOW;
	DWORD       dwMainExStyle = NULL;

	DWORD       dwScrollbarStyle = WS_DISABLED | WS_CHILD | WS_VISIBLE;

	DWORD       dwBoxStyle = BS_GROUPBOX | WS_CHILD | WS_VISIBLE;

	DWORD       dwGLVStyle = WS_DISABLED | WS_CHILD | WS_VISIBLE;
	DWORD       dwGLVExStyle = WS_EX_CLIENTEDGE;

	DWORD       dwLBStyle = LBS_NOTIFY | WS_VSCROLL | WS_CHILD | WS_DISABLED | WS_VISIBLE /*| LBS_DISABLENOSCROLL | LBS_WANTKEYBOARDINPUT*/;
	DWORD       dwLBExStyle = WS_EX_CLIENTEDGE;

	DWORD       dwButtonStyle = BS_PUSHBUTTON | WS_CHILD | WS_VISIBLE;
	DWORD       dwRadioStyle = BS_RADIOBUTTON | WS_CHILD | WS_VISIBLE;
	DWORD       dwCheckboxStyle = BS_AUTOCHECKBOX | WS_CHILD | WS_VISIBLE;

	HFONT       hFont;
	hFont = CreateFont(14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "MS Sans Serif");

///////////////////////////////////////////////////////////////////////////////////////////////////

	// register the main window class
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = (WNDPROC)WndProc_Main;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = ghinstApp;
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE +1);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = gszMainClassName;
	wc.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

	if (!RegisterClassEx(&wc))
		{
		MessageBox(NULL, "Window class registration failed for: Main window class.", "Error", MB_ICONEXCLAMATION | MB_APPLMODAL);
		return false;
		}

	// create the main menu
	ghmenuMain = CreateMenu();
	hmenuPopupFile = CreateMenu();
	hmenuPopupOptions = CreateMenu();

	AppendMenu(hmenuPopupFile,    MF_STRING,              ID_MENUOPEN,            "&Open...");
	AppendMenu(hmenuPopupFile,    MF_STRING | MF_GRAYED,  ID_MENUREVERT,          "&Revert");
	AppendMenu(hmenuPopupFile,    MF_SEPARATOR,           0,                      NULL);
	AppendMenu(hmenuPopupFile,    MF_STRING | MF_GRAYED,  ID_MENUSAVE,            "&Save");
	AppendMenu(hmenuPopupFile,    MF_STRING | MF_GRAYED,  ID_MENUSAVEAS,          "Save &as...");
	AppendMenu(hmenuPopupFile,    MF_SEPARATOR,           0,                      NULL);
	AppendMenu(hmenuPopupFile,    MF_STRING | MF_GRAYED,  ID_MENUCLOSE,           "&Close");

	AppendMenu(hmenuPopupOptions, MF_STRING,              ID_MENUOPTIONS,         "&Preferences...");
	AppendMenu(hmenuPopupOptions, MF_STRING,              ID_MENUHIDECONTROLS,    "&Toggle control panel (enter)");
	AppendMenu(hmenuPopupOptions, MF_STRING,              ID_MENUOFFSET,          "&Re-calculate character offset");

	AppendMenu(ghmenuMain,        MF_POPUP,               (UINT)hmenuPopupFile,   "&File");
	AppendMenu(ghmenuMain,        MF_POPUP,               (UINT)hmenuPopupOptions,"&Options");
	AppendMenu(ghmenuMain,        MF_STRING | MF_HELP,    ID_MENUABOUT,           "&About...");

	// create the main window
	ghwndMain = CreateWindowEx(dwMainExStyle, gszMainClassName, gszAppName, dwMainStyle, 256+2*SEP, SEP, 800-256-3*SEP, 570-SEP, NULL, ghmenuMain, ghinstApp, NULL);
	if (!ghwndMain) return false;
	SendMessage(ghwndMain, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

///////////////////////////////////////////////////////////////////////////////////////////////////

	// create scroll bars
	si.cbSize = sizeof(SCROLLINFO);
	si.fMask  = SIF_POS | SIF_RANGE;
	si.nPos   = 0;

	si.nMin   = -100;
	si.nMax   = 100;

	ghwndScrollPanY = CreateWindow("scrollbar", NULL, SBS_VERT | dwScrollbarStyle, 0, 0, 0, 0, ghwndMain, NULL, ghinstApp, NULL);
	if (!ghwndScrollPanY) return false;
	SetScrollInfo(ghwndScrollPanY, SB_CTL, &si, FALSE);

	ghwndScrollPanX = CreateWindow("scrollbar", NULL, SBS_HORZ | dwScrollbarStyle, 0, 0, 0, 0, ghwndMain, NULL, ghinstApp, NULL);
	if (!ghwndScrollPanX) return false;
	SetScrollInfo(ghwndScrollPanX, SB_CTL, &si, FALSE);

	ghwndScrollZoom = CreateWindow("scrollbar", NULL, SBS_VERT | dwScrollbarStyle, 0, 0, 0, 0, ghwndMain, NULL, ghinstApp, NULL);
	if (!ghwndScrollZoom) return false;
	SetScrollInfo(ghwndScrollZoom, SB_CTL, &si, FALSE);

	si.nMin   = -180;
	si.nMax   = 180;

	ghwndScrollPitch  = CreateWindow("scrollbar", NULL, SBS_VERT | dwScrollbarStyle, 0, 0, 0, 0, ghwndMain, NULL, ghinstApp, NULL);
	if (!ghwndScrollPitch) return false;
	SetScrollInfo(ghwndScrollPitch, SB_CTL, &si, FALSE);

	ghwndScrollYaw  = CreateWindow("scrollbar", NULL, SBS_HORZ | dwScrollbarStyle, 0, 0, 0, 0, ghwndMain, NULL, ghinstApp, NULL);
	if (!ghwndScrollYaw) return false;
	SetScrollInfo(ghwndScrollYaw, SB_CTL, &si, FALSE);

	// create mouselook button
	ghwndMouseLook  = CreateWindow("button", "M", dwButtonStyle | WS_DISABLED, 0, 0, 0, 0, ghwndMain, (HMENU)ID_MOUSELOOK, ghinstApp, NULL);
	if (!ghwndMouseLook) return false;
	SendMessage(ghwndMouseLook, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create reset view button
	ghwndResetView  = CreateWindow("button", "R", dwButtonStyle | WS_DISABLED, 0, 0, 0, 0, ghwndMain, (HMENU)ID_RESETVIEW, ghinstApp, NULL);
	if (!ghwndResetView) return false;
	SendMessage(ghwndResetView, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

///////////////////////////////////////////////////////////////////////////////////////////////////

	wc.lpfnWndProc = (WNDPROC)WndProc_Controls;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = gszControlsClassName;

	if (!RegisterClassEx(&wc))
		{ MessageBox(NULL, "Window class registration failed for: Controls window class.", "Error", MB_ICONEXCLAMATION | MB_APPLMODAL); return false; }

	// create the control panel
	ghwndControls = CreateWindowEx(NULL, gszControlsClassName, "Control Panel", WS_POPUP | WS_CAPTION | WS_SYSMENU, SEP, 4*SEP, 256, 464, ghwndMain, NULL, ghinstApp, NULL);
	if (!ghwndControls) return false;
	SendMessage(ghwndControls, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

///////////////////////////////////////////////////////////////////////////////////////////////////

	wc.lpfnWndProc = (WNDPROC)WndProc_Options;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = gszOptionsClassName;

	if (!RegisterClassEx(&wc))
		{ MessageBox(NULL, "Window class registration failed for: Options window class.", "Error", MB_ICONEXCLAMATION | MB_APPLMODAL); return false; }

	// create the options window
	ghwndOptions = CreateWindowEx(NULL, gszOptionsClassName, "Preferences", WS_POPUP | WS_CAPTION | WS_SYSMENU, 2*SEP, 5*SEP, 256, 216, ghwndMain, NULL, ghinstApp, NULL);
	if (!ghwndOptions) return false;
	SendMessage(ghwndOptions, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsBox1 = CreateWindow("button", "Texture loading priority", dwBoxStyle, 8, 8, 236, 60, ghwndOptions, NULL, ghinstApp, NULL);
	if (!ghwndOptionsBox1) return false;
	SendMessage(ghwndOptionsBox1, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsMatFirst = CreateWindow("button", "Try MAT first  (", dwRadioStyle, 16, 26, 88, 16, ghwndOptions, (HMENU)ID_OPTIONSMATFIRST, ghinstApp, NULL);
	if (!ghwndOptionsMatFirst) return false;
	SendMessage(ghwndOptionsMatFirst, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsBmpSecond = CreateWindow("button", "load BMP if MAT fails? )", dwCheckboxStyle, 106, 26, 134, 16, ghwndOptions, (HMENU)ID_OPTIONSBMPSECOND, ghinstApp, NULL);
	if (!ghwndOptionsBmpSecond) return false;
	SendMessage(ghwndOptionsBmpSecond, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsBmpFirst = CreateWindow("button", "Try BMP first  (", dwRadioStyle, 16, 46, 88, 16, ghwndOptions, (HMENU)ID_OPTIONSBMPFIRST, ghinstApp, NULL);
	if (!ghwndOptionsBmpFirst) return false;
	SendMessage(ghwndOptionsBmpFirst, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsMatSecond = CreateWindow("button", "load MAT if BMP fails? )", dwCheckboxStyle, 106, 46, 134, 16, ghwndOptions, (HMENU)ID_OPTIONSMATSECOND, ghinstApp, NULL);
	if (!ghwndOptionsMatSecond) return false;
	SendMessage(ghwndOptionsMatSecond, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsBox3 = CreateWindow("button", "CMP palette", dwBoxStyle, 8, 80, 236, 63, ghwndOptions, NULL, ghinstApp, NULL);
	if (!ghwndOptionsBox3) return false;
	SendMessage(ghwndOptionsBox3, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsCmpFile1 = CreateWindow("static", "File:", dwButtonStyle, 16, 98, 32, 16, ghwndOptions, NULL, ghinstApp, NULL);
	if (!ghwndOptionsCmpFile1) return false;
	SendMessage(ghwndOptionsCmpFile1, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsCmpFile2 = CreateWindow("static", "", dwButtonStyle, 42, 98, 192, 16, ghwndOptions, NULL, ghinstApp, NULL);
	if (!ghwndOptionsCmpFile2) return false;
	SendMessage(ghwndOptionsCmpFile2, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsCmpBrowse = CreateWindow("button", "browse...", dwButtonStyle, 16, 116, 64, 18, ghwndOptions, (HMENU)ID_OPTIONSCMPBROWSE, ghinstApp, NULL);
	if (!ghwndOptionsCmpBrowse) return false;
	SendMessage(ghwndOptionsCmpBrowse, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsDefault = CreateWindow("button", "default", dwButtonStyle, 80, 116, 64, 18, ghwndOptions, (HMENU)ID_OPTIONSCMPDEFAULT, ghinstApp, NULL);
	if (!ghwndOptionsDefault) return false;
	SendMessage(ghwndOptionsDefault, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsOK = CreateWindow("button", "ok", dwButtonStyle, 61, 159, 64, 18, ghwndOptions, (HMENU)ID_OPTIONSOK, ghinstApp, NULL);
	if (!ghwndOptionsOK) return false;
	SendMessage(ghwndOptionsOK, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	ghwndOptionsCancel = CreateWindow("button", "cancel", dwButtonStyle, 125, 159 , 64, 18, ghwndOptions, (HMENU)ID_OPTIONSCANCEL, ghinstApp, NULL);
	if (!ghwndOptionsCancel) return false;
	SendMessage(ghwndOptionsCancel, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

SendMessage(ghwndOptions, WM_UPDATE, 0, 0);

///////////////////////////////////////////////////////////////////////////////////////////////////

	// register an OpenGL compatible window class
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC | CS_SAVEBITS;
	wc.lpfnWndProc = (WNDPROC)WndProc_GLViewport;
	wc.hbrBackground = (HBRUSH)(COLOR_APPWORKSPACE+1);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = gszGLVClassName;

	if (!RegisterClassEx(&wc))
		{ MessageBox(NULL,"Window class registration failed for: OpenGL window class.", "Error", MB_ICONEXCLAMATION | MB_APPLMODAL); return false; }

	// create the GL Viewport (GLV)
	ghwndGLV = CreateWindowEx(dwGLVExStyle, gszGLVClassName, NULL, dwGLVStyle, 0, 0, 0, 0, ghwndMain, (HMENU)ID_GLV, ghinstApp, NULL);
	if (!ghwndGLV) return false;
	SendMessage(ghwndGLV, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

///////////////////////////////////////////////////////////////////////////////////////////////////

	// create bounding box
	ghwndMeshBox = CreateWindow("button", "Mesh", dwBoxStyle | WS_DISABLED, 8, 8, 118, 294, ghwndControls, NULL, ghinstApp, NULL);
	if (!ghwndMeshBox) return false;
	SendMessage(ghwndMeshBox, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create list box
	ghwndMeshLB = CreateWindowEx(dwLBExStyle, "listbox", NULL, dwLBStyle, 17, 23, 100, 240, ghwndControls, (HMENU)ID_MESHLB, ghinstApp, NULL);
	if (!ghwndMeshLB) return false;
	SendMessage(ghwndMeshLB, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create 'hide' button
	ghwndMeshHide = CreateWindow("button", "hide", dwButtonStyle | WS_DISABLED, 17, 256, 30, 18, ghwndControls, (HMENU)ID_MESHHIDE, ghinstApp, NULL);
	if (!ghwndMeshHide) return false;
	SendMessage(ghwndMeshHide, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create 'wire' button
	ghwndMeshWire = CreateWindow("button", "wire", dwButtonStyle | WS_DISABLED, 52, 256, 30, 18, ghwndControls, (HMENU)ID_MESHWIRE, ghinstApp, NULL);
	if (!ghwndMeshWire) return false;
	SendMessage(ghwndMeshWire, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create 'show' button
	ghwndMeshShow = CreateWindow("button", "show", dwButtonStyle | WS_DISABLED, 87, 256, 30, 18, ghwndControls, (HMENU)ID_MESHSHOW, ghinstApp, NULL);
	if (!ghwndMeshShow) return false;
	SendMessage(ghwndMeshShow, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create 'hide all' button
	ghwndMeshHideAll = CreateWindow("button", "all", dwButtonStyle | WS_DISABLED, 17, 274, 30, 18, ghwndControls, (HMENU)ID_MESHHIDEALL, ghinstApp, NULL);
	if (!ghwndMeshHideAll) return false;
	SendMessage(ghwndMeshHideAll, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create 'wire all' button
	ghwndMeshWireAll = CreateWindow("button", "all", dwButtonStyle | WS_DISABLED, 52, 274, 30, 18, ghwndControls, (HMENU)ID_MESHWIREALL, ghinstApp, NULL);
	if (!ghwndMeshWireAll) return false;
	SendMessage(ghwndMeshWireAll, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create 'show all' button
	ghwndMeshShowAll = CreateWindow("button", "all", dwButtonStyle | WS_DISABLED, 87, 274, 30, 18, ghwndControls, (HMENU)ID_MESHSHOWALL, ghinstApp, NULL);
	if (!ghwndMeshShowAll) return false;
	SendMessage(ghwndMeshShowAll, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

///////////////////////////////////////////////////////////////////////////////////////////////////

	// create bounding box
	ghwndGroupBox = CreateWindow("button", "Texture", dwBoxStyle | WS_DISABLED, 124, 8, 118, 294, ghwndControls, NULL, ghinstApp, NULL);
	if (!ghwndGroupBox) return false;
	SendMessage(ghwndGroupBox, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create list box
	ghwndGroupLB = CreateWindowEx(dwLBExStyle, "listbox", NULL, dwLBStyle, 133, 23, 100, 240, ghwndControls, (HMENU)ID_GROUPLB, ghinstApp, NULL);
	if (!ghwndGroupLB) return false;
	SendMessage(ghwndGroupLB, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create 'select none' button
	ghwndGroupSelectNone = CreateWindow("button", "sel none", dwButtonStyle | WS_DISABLED, 133, 256, 100, 18, ghwndControls, (HMENU)ID_GROUPSELECTNONE, ghinstApp, NULL);
	if (!ghwndGroupSelectNone) return false;
	SendMessage(ghwndGroupSelectNone, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create 'change bmp' button
	ghwndGroupChangeBmp = CreateWindow("button", "change ...", dwButtonStyle | WS_DISABLED, 133, 274, 100, 18, ghwndControls, (HMENU)ID_GROUPCHANGEBMP, ghinstApp, NULL);
	if (!ghwndGroupChangeBmp) return false;
	SendMessage(ghwndGroupChangeBmp, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

///////////////////////////////////////////////////////////////////////////////////////////////////

	// create bounding box
	ghwndUVBox = CreateWindow("button", "Auto remap", dwBoxStyle | WS_DISABLED, 8, 307, 234, 44, ghwndControls, NULL, ghinstApp, NULL);
	if (!ghwndUVBox) return false;
	SendMessage(ghwndUVBox, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'ortho front' button
	ghwndUVOrthoFront = CreateWindow("button", "front", dwButtonStyle | WS_DISABLED, 17, 323, 36, 18, ghwndControls, (HMENU)ID_UVORTHOFRONT, ghinstApp, NULL);
	if (!ghwndUVOrthoFront) return false;
	SendMessage(ghwndUVOrthoFront, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'ortho back' button
	ghwndUVOrthoBack = CreateWindow("button", "back", dwButtonStyle | WS_DISABLED, 53, 323, 36, 18, ghwndControls, (HMENU)ID_UVORTHOBACK, ghinstApp, NULL);
	if (!ghwndUVOrthoBack) return false;
	SendMessage(ghwndUVOrthoBack, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'ortho top' button
	ghwndUVOrthoTop = CreateWindow("button", "top", dwButtonStyle | WS_DISABLED, 89, 323, 36, 18, ghwndControls, (HMENU)ID_UVORTHOTOP, ghinstApp, NULL);
	if (!ghwndUVOrthoTop) return false;
	SendMessage(ghwndUVOrthoTop, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'ortho bottom' button
	ghwndUVOrthoBottom = CreateWindow("button", "bott", dwButtonStyle | WS_DISABLED, 125, 323, 36, 18, ghwndControls, (HMENU)ID_UVORTHOBOTTOM, ghinstApp, NULL);
	if (!ghwndUVOrthoBottom) return false;
	SendMessage(ghwndUVOrthoBottom, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'ortho left' button
	ghwndUVOrthoLeft = CreateWindow("button", "left", dwButtonStyle | WS_DISABLED, 161, 323, 36, 18, ghwndControls, (HMENU)ID_UVORTHOLEFT, ghinstApp, NULL);
	if (!ghwndUVOrthoLeft) return false;
	SendMessage(ghwndUVOrthoLeft, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'ortho right' button
	ghwndUVOrthoRight = CreateWindow("button", "right", dwButtonStyle | WS_DISABLED, 197, 323, 36, 18, ghwndControls, (HMENU)ID_UVORTHORIGHT, ghinstApp, NULL);
	if (!ghwndUVOrthoRight) return false;
	SendMessage(ghwndUVOrthoRight, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create bounding box
	ghwndUVBox2 = CreateWindow("button", "Manual remap", dwBoxStyle | WS_DISABLED, 8, 356, 234, 68, ghwndControls, NULL, ghinstApp, NULL);
	if (!ghwndUVBox2) return false;
	SendMessage(ghwndUVBox2, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'edit' button
	ghwndUVEdit = CreateWindow("button", "edit", dwButtonStyle | WS_DISABLED, 17, 372, 36, 18, ghwndControls, (HMENU)ID_UVEDIT, ghinstApp, NULL);
	if (!ghwndUVEdit) return false;
	SendMessage(ghwndUVEdit, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'return' button
	ghwndUVHelp = CreateWindow("button", "help", dwButtonStyle | WS_DISABLED, 53, 372, 36, 18, ghwndControls, (HMENU)ID_UVHELP, ghinstApp, NULL);
	if (!ghwndUVHelp) return false;
	SendMessage(ghwndUVHelp, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'select:' static text
	ghwndUVStaticSel = CreateWindow("static", "sel:", dwButtonStyle | WS_DISABLED, 104, 374, 20, 18, ghwndControls, NULL, ghinstApp, NULL);
	if (!ghwndUVStaticSel) return false;
	SendMessage(ghwndUVStaticSel, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'select all' button
	ghwndUVSelectAll = CreateWindow("button", "all", dwButtonStyle | WS_DISABLED, 125, 372, 36, 18, ghwndControls, (HMENU)ID_UVSELECTALL, ghinstApp, NULL);
	if (!ghwndUVSelectAll) return false;
	SendMessage(ghwndUVSelectAll, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'select none' button
	ghwndUVSelectNone = CreateWindow("button", "none", dwButtonStyle | WS_DISABLED, 161, 372, 36, 18, ghwndControls, (HMENU)ID_UVSELECTNONE, ghinstApp, NULL);
	if (!ghwndUVSelectNone) return false;
	SendMessage(ghwndUVSelectNone, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// create 'select invert' button
	ghwndUVSelectInvert = CreateWindow("button", "inv", dwButtonStyle | WS_DISABLED, 197, 372, 36, 18, ghwndControls, (HMENU)ID_UVSELECTINVERT, ghinstApp, NULL);
	if (!ghwndUVSelectInvert) return false;
	SendMessage(ghwndUVSelectInvert, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));

	// create 'scale' button
	ghwndUVScale = CreateWindow("button", "scale", dwButtonStyle | WS_DISABLED, 17, 396, 36, 18, ghwndControls, (HMENU)ID_UVSCALE, ghinstApp, NULL);
	if (!ghwndUVScale) return false;
	SendMessage(ghwndUVScale, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// amount
	ghwndUVScaleAmount = CreateWindowEx(WS_EX_CLIENTEDGE, "Edit", NULL, WS_DISABLED | WS_CHILD | WS_VISIBLE | ES_AUTOHSCROLL | ES_LEFT, 52, 396, 37, 18, ghwndControls, (HMENU)ID_UVAMOUNT, ghinstApp, NULL);
	if (!ghwndUVScaleAmount) return false;
	SendMessage(ghwndUVScaleAmount, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// on
	ghwndUVStaticOn = CreateWindow("static", "on:", dwButtonStyle | WS_DISABLED, 104, 398, 20, 18, ghwndControls, NULL, ghinstApp, NULL);
	if (!ghwndUVStaticOn) return false;
	SendMessage(ghwndUVStaticOn, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// X
	ghwndUVScaleX = CreateWindow("button", "x", dwRadioStyle | WS_DISABLED, 125, 398, 36, 18, ghwndControls, (HMENU)ID_UVXAXIS, ghinstApp, NULL);
	if (!ghwndUVScaleX) return false;
	SendMessage(ghwndUVScaleX, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// Y
	ghwndUVScaleY = CreateWindow("button", "y", dwRadioStyle | WS_DISABLED, 161, 398, 36, 18, ghwndControls, (HMENU)ID_UVYAXIS, ghinstApp, NULL);
	if (!ghwndUVScaleY) return false;
	SendMessage(ghwndUVScaleY, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	// BOTH
	ghwndUVScaleBoth = CreateWindow("button", "xy", dwRadioStyle | WS_DISABLED, 197, 398, 36, 18, ghwndControls, (HMENU)ID_UVBOTHAXES, ghinstApp, NULL);
	if (!ghwndUVScaleBoth) return false;
	SendMessage(ghwndUVScaleBoth, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(true, 0));
	SendMessage(ghwndUVScaleBoth, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);

///////////////////////////////////////////////////////////////////////////////////////////////////

	/*// retrieve client area dimensions
	if (!(GetClientRect(ghwndMain, &grectMain))) return false;*/

	ShowWindow(ghwndControls, SW_NORMAL);
	ShowWindow(ghwndMain, SW_NORMAL);

	// initialize the GLV and OpenGL settings
	ghdcGLV = GetDC(ghwndGLV);
	GL_InitializePixelFormat(ghdcGLV);
	ghrcGLV = wglCreateContext(ghdcGLV);
	wglMakeCurrent(ghdcGLV, ghrcGLV);
	GL_InitializeSettings();
	if (!(GetClientRect(ghwndGLV, &grectGLV))) return false;
	GL_InitializeOrthoView();

	return true;
	}

void Menu_Enable(bool flag)
	{
	EnableMenuItem(ghmenuMain, ID_MENUOPEN,   MF_ENABLED);

	if (flag)
		{
		EnableMenuItem(ghmenuMain, ID_MENUREVERT, MF_ENABLED);
		EnableMenuItem(ghmenuMain, ID_MENUCLOSE,  MF_ENABLED);
		EnableMenuItem(ghmenuMain, ID_MENUSAVE,   MF_ENABLED);
		EnableMenuItem(ghmenuMain, ID_MENUSAVEAS, MF_ENABLED);
		}
	else
		{
		EnableMenuItem(ghmenuMain, ID_MENUREVERT, MF_DISABLED | MF_GRAYED);
		EnableMenuItem(ghmenuMain, ID_MENUCLOSE,  MF_DISABLED | MF_GRAYED);
		EnableMenuItem(ghmenuMain, ID_MENUSAVE,   MF_DISABLED | MF_GRAYED);
		EnableMenuItem(ghmenuMain, ID_MENUSAVEAS, MF_DISABLED | MF_GRAYED);
		}
	}

void ProgramMode1(void)
	{
	giProgramMode = 1;

	EnableWindow(ghwndMeshBox, false);
	EnableWindow(ghwndMeshLB, false);
	EnableWindow(ghwndMeshHide, false);
	EnableWindow(ghwndMeshWire, false);
	EnableWindow(ghwndMeshShow, false);
	EnableWindow(ghwndMeshHideAll, false);
	EnableWindow(ghwndMeshWireAll, false);
	EnableWindow(ghwndMeshShowAll, false);

	EnableWindow(ghwndGroupBox, false);
	EnableWindow(ghwndGroupLB, false);
	EnableWindow(ghwndGroupChangeBmp, false);
	EnableWindow(ghwndGroupSelectNone, false);

	EnableWindow(ghwndUVBox, false);
	EnableWindow(ghwndUVEdit, false);
	EnableWindow(ghwndUVHelp, false);
	EnableWindow(ghwndUVStaticOn, false);
	EnableWindow(ghwndUVSelectAll, false);
	EnableWindow(ghwndUVSelectNone, false);
	EnableWindow(ghwndUVSelectInvert, false);

	EnableWindow(ghwndUVOrthoFront, false);
	EnableWindow(ghwndUVOrthoBack, false);
	EnableWindow(ghwndUVOrthoTop, false);
	EnableWindow(ghwndUVOrthoBottom, false);
	EnableWindow(ghwndUVOrthoLeft, false);
	EnableWindow(ghwndUVOrthoRight, false);

	EnableWindow(ghwndGLV, false);
	EnableWindow(ghwndScrollPitch, false);
	EnableWindow(ghwndScrollYaw, false);
	EnableWindow(ghwndScrollPanX, false);
	EnableWindow(ghwndScrollPanY, false);
	EnableWindow(ghwndScrollZoom, false);
	EnableWindow(ghwndMouseLook, false);
	EnableWindow(ghwndResetView, false);
	}

void ProgramMode1to2(void)
	{
	giProgramMode = 2;

	EnableWindow(ghwndMeshBox, true);
	EnableWindow(ghwndMeshLB, true);
	EnableWindow(ghwndMeshHide, true);
	EnableWindow(ghwndMeshWire, true);
	EnableWindow(ghwndMeshShow, true);
	EnableWindow(ghwndMeshHideAll, true);
	EnableWindow(ghwndMeshWireAll, true);
	EnableWindow(ghwndMeshShowAll, true);

	EnableWindow(ghwndGroupBox, true);
	EnableWindow(ghwndGroupLB, true);

	EnableWindow(ghwndGLV, true);
	EnableWindow(ghwndScrollPitch, true);
	EnableWindow(ghwndScrollYaw, true);
	EnableWindow(ghwndScrollPanX, true);
	EnableWindow(ghwndScrollPanY, true);
	EnableWindow(ghwndScrollZoom, true);
	EnableWindow(ghwndMouseLook, true);
	EnableWindow(ghwndResetView, true);
	}

void ProgramMode2to3(bool flag)
	{
	if(flag) giProgramMode = 3;
	else giProgramMode = 2;

	EnableWindow(ghwndGroupChangeBmp, flag);
	EnableWindow(ghwndGroupSelectNone, flag);

	EnableWindow(ghwndUVBox, flag);
	EnableWindow(ghwndUVBox2, flag);

	EnableWindow(ghwndUVEdit, flag);

	EnableWindow(ghwndUVOrthoFront, flag);
	EnableWindow(ghwndUVOrthoBack, flag);
	EnableWindow(ghwndUVOrthoTop, flag);
	EnableWindow(ghwndUVOrthoBottom, flag);
	EnableWindow(ghwndUVOrthoLeft, flag);
	EnableWindow(ghwndUVOrthoRight, flag);
	}

void ProgramMode3to4(bool flag)
	{
	if (flag) giProgramMode = 4;
	else giProgramMode = 3;

	EnableWindow(ghwndMeshLB, !flag);
	EnableWindow(ghwndMeshHide, !flag);
	EnableWindow(ghwndMeshWire, !flag);
	EnableWindow(ghwndMeshShow, !flag);
	EnableWindow(ghwndMeshHideAll, !flag);
	EnableWindow(ghwndMeshWireAll, !flag);
	EnableWindow(ghwndMeshShowAll, !flag);

	EnableWindow(ghwndGroupLB, !flag);
	EnableWindow(ghwndGroupChangeBmp, !flag);
	EnableWindow(ghwndGroupSelectNone, !flag);

	if (flag) SetWindowText(ghwndUVEdit, "return");
	else SetWindowText(ghwndUVEdit, "edit");

	EnableWindow(ghwndUVHelp, flag);
	EnableWindow(ghwndUVStaticSel, flag);
	EnableWindow(ghwndUVSelectAll, flag);
	EnableWindow(ghwndUVSelectNone, flag);
	EnableWindow(ghwndUVSelectInvert, flag);
	EnableWindow(ghwndUVScale, flag);
	EnableWindow(ghwndUVScaleAmount, flag);
	EnableWindow(ghwndUVStaticOn, flag);
	EnableWindow(ghwndUVScaleX, flag);
	EnableWindow(ghwndUVScaleY, flag);
	EnableWindow(ghwndUVScaleBoth, flag);

	EnableWindow(ghwndScrollPitch, !flag);
	EnableWindow(ghwndScrollYaw, !flag);
	EnableWindow(ghwndScrollPanX, !flag);
	EnableWindow(ghwndScrollPanY, !flag);
	EnableWindow(ghwndScrollZoom, !flag);
	EnableWindow(ghwndMouseLook, !flag);
	EnableWindow(ghwndResetView, !flag);
	}

///////////////////////////////////////////////////////////////////////////////////////////////////

void AllWindows_Close(void)
	{
	SendMessage(ghwndMeshBox, WM_CLOSE, 0, 0);
	SendMessage(ghwndMeshLB, WM_CLOSE, 0, 0);
	SendMessage(ghwndMeshHide, WM_CLOSE, 0, 0);
	SendMessage(ghwndMeshWire, WM_CLOSE, 0, 0);
	SendMessage(ghwndMeshShow, WM_CLOSE, 0, 0);
	SendMessage(ghwndMeshHideAll, WM_CLOSE, 0, 0);
	SendMessage(ghwndMeshWireAll, WM_CLOSE, 0, 0);
	SendMessage(ghwndMeshShowAll, WM_CLOSE, 0, 0);

	SendMessage(ghwndGroupBox, WM_CLOSE, 0, 0);
	SendMessage(ghwndGroupLB, WM_CLOSE, 0, 0);
	SendMessage(ghwndGroupChangeBmp, WM_CLOSE, 0, 0);
	SendMessage(ghwndGroupSelectNone, WM_CLOSE, 0, 0);

	SendMessage(ghwndUVBox, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVBox2, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVOrthoFront, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVOrthoBack, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVOrthoTop, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVOrthoBottom, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVOrthoLeft, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVOrthoRight, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVEdit, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVHelp, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVStaticSel, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVSelectAll, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVSelectNone, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVSelectInvert, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVScale, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVScaleAmount, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVStaticOn, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVScaleX, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVScaleY, WM_CLOSE, 0, 0);
	SendMessage(ghwndUVScaleBoth, WM_CLOSE, 0, 0);

	SendMessage(ghwndGLV, WM_CLOSE, 0, 0);
	SendMessage(ghwndScrollPitch, WM_CLOSE, 0, 0);
	SendMessage(ghwndScrollYaw, WM_CLOSE, 0, 0);
	SendMessage(ghwndScrollPanX, WM_CLOSE, 0, 0);
	SendMessage(ghwndScrollPanY, WM_CLOSE, 0, 0);
	SendMessage(ghwndScrollZoom, WM_CLOSE, 0, 0);
	SendMessage(ghwndMouseLook, WM_CLOSE, 0, 0);
	SendMessage(ghwndResetView, WM_CLOSE, 0, 0);

	SendMessage(ghwndControls, WM_USERCLOSE, 0, 0);

	SendMessage(ghwndOptions, WM_USERCLOSE, 0, 0);
	}

void Command_Open(char *szPathname)
	{
	int i;

	gmodel = new C3DOModel;

	if (!(gmodel->Load_File(szPathname)))
		{
		MessageBox(ghwndMain, "Error loading file.", "File error", MB_APPLMODAL);
		delete gmodel;
		gmodel = NULL;
		}
	else
		{
		Support_SetGlobalFilenameInfo(szPathname);
		Support_SetAppCaption();

		if (!(gmodel->Initialize_Model()))
			{
			MessageBox(ghwndMain, "Error initializing model.", "Error", MB_APPLMODAL);
			delete gmodel;
			gmodel = NULL;

			Support_SetGlobalFilenameInfo(gszEmptyString);
			Support_SetAppCaption();
			}
		else
			{
			ProgramMode1to2();

			for(i = 0; i < gmodel->numMeshes; i++)
				SendMessage(ghwndMeshLB, LB_INSERTSTRING, i, (LPARAM)gmodel->meshes[i].jkName);
			PostMessage(ghwndMeshLB, WM_PAINT, 0, 0);
			SendMessage(ghwndMeshLB, LB_SETCURSEL, 0, 0);

			for(i = 0; i < gmodel->meshes[0].numglGroups; i++)
				SendMessage(ghwndGroupLB, LB_INSERTSTRING, i, (LPARAM)gmodel->meshes[0].glGroups[i].glGroupName);
			PostMessage(ghwndGroupLB, WM_PAINT, 0, 0);

			Menu_Enable(true);

			gbModelMode_Active = true;

			GL_InitializeOrthoView();

			gCamera->ResetView(gmodel->radius);
			}
		}
	}

void Command_Close(void)
	{
	int m;

	gbModelMode_Active = false;
	gbUVMapMode_Active = false;

	SendMessage(ghwndMeshLB, LB_RESETCONTENT, 0, 0);
	SendMessage(ghwndGroupLB, LB_RESETCONTENT, 0, 0);

	ProgramMode1();
	Menu_Enable(false);

	Support_SetGlobalFilenameInfo(gszEmptyString);
	Support_SetAppCaption();

	for(m = 0; m < gmodel->numMaterials; m++)
		glDeleteTextures(1, gmodel->materials[m].glTexture);

	if (gmodel != NULL)
		delete gmodel;
	gmodel = NULL;

	GL_ClearGLV();
	}

void Command_Save(void)
	{
	gmodel->Save_File(gszProjectPathname);
	}

void Command_SaveAs(char *szPathname)
	{
	gmodel->Save_File(szPathname);

	Support_SetGlobalFilenameInfo(szPathname);
	Support_SetAppCaption();
	}

void Support_SetGlobalFilenameInfo(char *szPathname)
	{
	int i, titlePos;

	if (szPathname[0] != '\0')
		{
		// set the global path name
		strcpy(gszProjectPathname, szPathname);

		// set the global project folder
		strcpy(gszProjectFolder, szPathname);
		titlePos = strlen(gszProjectFolder);
		while(gszProjectFolder[--titlePos] != '\\');
		gszProjectFolder[++titlePos] = '\0';

		// set the global title name
		i = 0;
		while(szPathname[titlePos] != '\0')
			gszProjectTitlename[i++] = szPathname[titlePos++];
		gszProjectTitlename[i] = '\0';
		}

	else
		{
		strcpy(gszProjectPathname,  gszEmptyString);
		strcpy(gszProjectFolder,    gszEmptyString);
		strcpy(gszProjectTitlename, gszEmptyString);
		}
	}

void Support_SetAppCaption()
	{
	char caption[32 + _MAX_FNAME];

	if (gszProjectTitlename[0] != '\0')
		{
		strcpy(caption, gszAppName);
		strcat(caption, " - ");
		strcat(caption, gszProjectTitlename);
		SetWindowText(ghwndMain, caption);
		}

	else
		SetWindowText(ghwndMain, gszAppName);
	}

void Options_LoadFromReg()
	{
	HKEY hKey;

	unsigned long disposition;
	unsigned char szMat[] = "MAT\0";
	unsigned char szBoth[] = "BOTH\0";

	unsigned long type;
	unsigned long length;
	unsigned char szPrimary[_MAX_PATH];
	unsigned char szSecondary[_MAX_PATH];

	RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\ForceRemap\\",
	  0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
	  &hKey, &disposition);

	if (disposition == REG_CREATED_NEW_KEY)
		{
		RegSetValueEx(hKey, "CmpFile", 0, REG_SZ, gszDfltCmp, sizeof(gszDfltCmp));
		RegSetValueEx(hKey, "PrimaryTex", 0, REG_SZ, szMat, sizeof(szMat));
		RegSetValueEx(hKey, "SecondaryTex", 0, REG_SZ, szBoth, sizeof(szBoth));
		}

	length = _MAX_PATH;
	if (ERROR_SUCCESS == RegQueryValueEx(hKey, "PrimaryTex", NULL, &type, szPrimary, &length))
		{
		if (!strcmp((char*)szPrimary, "BMP"))
			{
			gMatFirst = false;
			gBmpFirst = true;
			}
		else
			{
			gMatFirst = true;
			gBmpFirst = false;
			}
		}
	else
		{
		gMatFirst = true;
		gBmpFirst = false;
		}

	length = _MAX_PATH;
	if (ERROR_SUCCESS == RegQueryValueEx(hKey, "SecondaryTex", NULL, &type, szSecondary, &length))
		{
		if (!strcmp((char*)szSecondary, "NONE"))
			{
			gMatSecond = false;
			gBmpSecond = false;
			}
		else if (!strcmp((char*)szSecondary, "MAT"))
			{
			gMatSecond = true;
			gBmpSecond = false;
			}
		else if (!strcmp((char*)szSecondary, "BMP"))
			{
			gMatSecond = false;
			gBmpSecond = true;
			}
		else
			{
			gMatSecond = true;
			gBmpSecond = true;
			}
		}
	else
		{
		gMatSecond = true;
		gBmpSecond = true;
		}

	length = _MAX_PATH;
	if (ERROR_SUCCESS == RegQueryValueEx(hKey, "CmpFile", NULL, &type, gCmpPath, &length))
		{
		ifstream cmpfile;
		cmpfile.open((char*)gCmpPath, ios::nocreate, filebuf::sh_none);
		if (cmpfile.fail())
			strcpy((char*)gCmpPath, (char*)gszDfltCmp);
		else
			cmpfile.close();
		}
	else
		{
		strcpy((char*)gCmpPath, (char*)gszDfltCmp);
		}
	LoadCmpFile();

	RegCloseKey(hKey);
	}

void Options_SaveToReg()
	{
	HKEY hKey;
	unsigned long disposition;

	unsigned char szNone[] = "NONE\0";
	unsigned char szMat [] = "MAT\0";
	unsigned char szBmp [] = "BMP\0";
	unsigned char szBoth[] = "BOTH\0";


	RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\ForceRemap\\",
	  0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
	  &hKey, &disposition);

	RegSetValueEx(hKey, "CmpFile", 0, REG_SZ, gCmpPath, sizeof(gCmpPath));

	if (!gMatFirst)
		RegSetValueEx(hKey, "PrimaryTex", 0, REG_SZ, szBmp, sizeof(szBmp));
	else
		RegSetValueEx(hKey, "PrimaryTex", 0, REG_SZ, szMat, sizeof(szMat));


	if (!gMatSecond && !gBmpSecond)
		RegSetValueEx(hKey, "SecondaryTex", 0, REG_SZ, szNone, sizeof(szNone));
	else if (gMatSecond && !gBmpSecond)
		RegSetValueEx(hKey, "SecondaryTex", 0, REG_SZ, szMat, sizeof(szMat));
	else if (!gMatSecond && gBmpSecond)
		RegSetValueEx(hKey, "SecondaryTex", 0, REG_SZ, szBmp, sizeof(szBmp));
	else
		RegSetValueEx(hKey, "SecondaryTex", 0, REG_SZ, szBoth, sizeof(szBoth));

	RegCloseKey(hKey);
	}

void Msg(HWND hwnd, char *szCaption, char *szFormat, ...)
	{
	char szBuffer[256];
	char *pArguments;

	pArguments = (char *) &szFormat + sizeof(szFormat);
	vsprintf(szBuffer, szFormat, pArguments);
	MessageBox(hwnd, szBuffer, szCaption, MB_APPLMODAL);
	}
